import numpy as np
import pytest
from astropy import cosmology as cosmo

from autolens.data import ccd as im, convolution
from autolens.data.array import grids
from autolens.data.array import mask as msk
from autolens.data.array import scaled_array
from autolens.lens import ray_tracing
from autolens.lens.util import lens_fit_util as util
from autolens.model.galaxy import galaxy as g
from autolens.model.galaxy.util import galaxy_util
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg
from autolens.model.profiles import light_profiles as lp
from test.unit.mock.mock_galaxy import MockHyperGalaxy


@pytest.fixture(name='mask')
def make_mask():
    return msk.Mask(array=np.array([[True, True, True, True],
                                    [True, False, False, True],
                                    [True, False, False, True],
                                    [True, True, True, True]]), pixel_scale=1.0)


@pytest.fixture(name='blurring_mask')
def make_blurring_mask():
    return msk.Mask(array=np.array([[False, False, False, False],
                                    [False, True, True, False],
                                    [False, True, True, False],
                                    [False, False, False, False]]), pixel_scale=1.0)


@pytest.fixture(name='convolver_no_blur')
def make_convolver_no_blur(mask, blurring_mask):
    psf = np.array([[0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0]])

    return convolution.ConvolverImage(mask=mask, blurring_mask=blurring_mask, psf=psf)


@pytest.fixture(name='convolver_blur')
def make_convolver_blur(mask, blurring_mask):
    psf = np.array([[1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0]])

    return convolution.ConvolverImage(mask=mask, blurring_mask=blurring_mask, psf=psf)


@pytest.fixture(name="galaxy_light")
def make_galaxy_light():
    return g.Galaxy(light_profile=lp.EllipticalSersic(centre=(0.1, 0.1), axis_ratio=1.0, phi=0.0, intensity=1.0,
                                                      effective_radius=0.6, sersic_index=4.0))


class TestBlurImages:

    def test__2x2_image_all_1s__3x3__psf_central_1__no_blurring(self, convolver_no_blur):
        image_1d = np.array([1.0, 1.0, 1.0, 1.0])
        blurring_image_1d = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])

        blurred_image_1d = util.blurred_image_1d_from_1d_unblurred_and_blurring_images(unblurred_image_1d=image_1d,
                                                                                       blurring_image_1d=blurring_image_1d,
                                                                                       convolver=convolver_no_blur)

        assert (blurred_image_1d == np.array([1.0, 1.0, 1.0, 1.0])).all()

    def test__2x2_image_all_1s__3x3_psf_all_1s__image_blurs_to_4s(self, convolver_blur):
        image_1d = np.array([1.0, 1.0, 1.0, 1.0])
        blurring_image_1d = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])

        blurred_image_1d = util.blurred_image_1d_from_1d_unblurred_and_blurring_images(unblurred_image_1d=image_1d,
                                                                                       blurring_image_1d=blurring_image_1d,
                                                                                       convolver=convolver_blur)
        assert (blurred_image_1d == np.array([4.0, 4.0, 4.0, 4.0])).all()

    def test__2x2_unblurred_image_all_1s__3x3_psf_no_blurring__image_blurs_to_itself(self, convolver_no_blur):
        unblurred_image_1d = np.array([1.0, 1.0, 1.0, 1.0])
        blurring_image_1d = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])

        blurred_image = util.blurred_image_1d_from_1d_unblurred_and_blurring_images(
            unblurred_image_1d=unblurred_image_1d, blurring_image_1d=blurring_image_1d, convolver=convolver_no_blur)

        assert (blurred_image == np.array([1.0, 1.0, 1.0, 1.0])).all()

    def test__2x2_unblurred_image_all_1s__3x3_psf_all_1s__image_blurs_to_9s(self, convolver_blur):
        unblurred_image_1d = np.array([1.0, 1.0, 1.0, 1.0])
        blurring_image_1d = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])

        blurred_image = util.blurred_image_1d_from_1d_unblurred_and_blurring_images(
            unblurred_image_1d=unblurred_image_1d, blurring_image_1d=blurring_image_1d, convolver=convolver_blur)

        assert (blurred_image == np.array([9.0, 9.0, 9.0, 9.0])).all()

    def test__2x2_image__psf_is_non_symmetric_producing_l_shape(self, mask, blurring_mask):
        psf = np.array([[0.0, 3.0, 0.0],
                        [0.0, 2.0, 1.0],
                        [0.0, 0.0, 0.0]])

        convolver = convolution.ConvolverImage(mask=mask, blurring_mask=blurring_mask, psf=psf)

        unblurred_image_1d = np.array([1.0, 2.0, 3.0, 4.0])
        blurring_image_1d = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0])

        blurred_image = util.blurred_image_1d_from_1d_unblurred_and_blurring_images(
            unblurred_image_1d=unblurred_image_1d, blurring_image_1d=blurring_image_1d, convolver=convolver)

        # Manually compute result of convolution, which is each central value *2.0 plus its 2 appropriate neighbors

        blurred_image_manual_0 = 2.0 * unblurred_image_1d[0] + 3.0 * unblurred_image_1d[2] + blurring_image_1d[4]
        blurred_image_manual_1 = 2.0 * unblurred_image_1d[1] + 3.0 * unblurred_image_1d[3] + unblurred_image_1d[0]
        blurred_image_manual_2 = 2.0 * unblurred_image_1d[2] + 3.0 * blurring_image_1d[9] + blurring_image_1d[6]
        blurred_image_manual_3 = 2.0 * unblurred_image_1d[3] + 3.0 * blurring_image_1d[10] + unblurred_image_1d[2]

        assert blurred_image_manual_0 == pytest.approx(blurred_image[0], 1e-6)
        assert blurred_image_manual_1 == pytest.approx(blurred_image[1], 1e-6)
        assert blurred_image_manual_2 == pytest.approx(blurred_image[2], 1e-6)
        assert blurred_image_manual_3 == pytest.approx(blurred_image[3], 1e-6)


class TestInversionEvidence:

    def test__simple_values(self):
        likelihood_with_regularization_terms = \
            util.likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
                chi_squared=3.0, regularization_term=6.0, noise_normalization=2.0)

        assert likelihood_with_regularization_terms == -0.5 * (3.0 + 6.0 + 2.0)

        evidences = util.evidence_from_inversion_terms(chi_squared=3.0, regularization_term=6.0,
                                                       log_curvature_regularization_term=9.0,
                                                       log_regularization_term=10.0, noise_normalization=30.0)

        assert evidences == -0.5 * (3.0 + 6.0 + 9.0 - 10.0 + 30.0)


class TestBlurredImageOfPlanes:

    def test__blurred_image_of_planes__real_tracer__2x2_image__psf_is_non_symmetric_producing_l_shape(self, mask,
                                                                                                      convolver_blur):
        data_grid_stack = grids.GridStack.grid_stack_from_mask_sub_grid_size_and_psf_shape(mask=mask, sub_grid_size=1,
                                                                                           psf_shape=(3, 3))

        g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))
        g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=2.0))

        tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0], source_galaxies=[g1],
                                                     image_plane_grid_stack=data_grid_stack)

        blurred_lens_image = convolver_blur.convolve_image(image_array=tracer.image_plane.image_plane_image_1d,
                                                           blurring_array=tracer.image_plane.image_plane_blurring_image_1d)
        blurred_lens_image = data_grid_stack.regular.scaled_array_2d_from_array_1d(array_1d=blurred_lens_image)

        blurred_source_image = convolver_blur.convolve_image(image_array=tracer.source_plane.image_plane_image_1d,
                                                             blurring_array=tracer.source_plane.image_plane_blurring_image_1d)
        blurred_source_image = data_grid_stack.regular.scaled_array_2d_from_array_1d(array_1d=blurred_source_image)

        blurred_image_of_planes = util.blurred_image_of_planes_from_1d_images_and_convolver(
            total_planes=tracer.total_planes, image_plane_image_1d_of_planes=tracer.image_plane_image_1d_of_planes,
            image_plane_blurring_image_1d_of_planes=tracer.image_plane_blurring_image_1d_of_planes,
            convolver=convolver_blur, map_to_scaled_array=data_grid_stack.regular.array_2d_from_array_1d)

        assert (blurred_image_of_planes[0] == blurred_lens_image).all()
        assert (blurred_image_of_planes[1] == blurred_source_image).all()

    def test__same_as_above_but_multi_tracer(self, mask, convolver_blur):
        data_grid_stack = grids.GridStack.grid_stack_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                           sub_grid_size=1,
                                                                                           psf_shape=(3, 3))

        g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0), redshift=0.1)
        g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=2.0), redshift=0.2)
        g2 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=3.0), redshift=0.3)

        tracer = ray_tracing.TracerMultiPlanes(galaxies=[g0, g1, g2], image_plane_grid_stack=data_grid_stack,
                                               cosmology=cosmo.Planck15)

        blurred_plane_image_0 = convolver_blur.convolve_image(tracer.planes[0].image_plane_image_1d,
                                                              tracer.planes[0].image_plane_blurring_image_1d)

        blurred_plane_image_1 = convolver_blur.convolve_image(tracer.planes[1].image_plane_image_1d,
                                                              tracer.planes[1].image_plane_blurring_image_1d)

        blurred_plane_image_2 = convolver_blur.convolve_image(tracer.planes[2].image_plane_image_1d,
                                                              tracer.planes[2].image_plane_blurring_image_1d)

        blurred_plane_image_0 = data_grid_stack.regular.scaled_array_2d_from_array_1d(blurred_plane_image_0)
        blurred_plane_image_1 = data_grid_stack.regular.scaled_array_2d_from_array_1d(blurred_plane_image_1)
        blurred_plane_image_2 = data_grid_stack.regular.scaled_array_2d_from_array_1d(blurred_plane_image_2)

        blurred_image_of_planes = util.blurred_image_of_planes_from_1d_images_and_convolver(
            total_planes=tracer.total_planes, image_plane_image_1d_of_planes=tracer.image_plane_image_1d_of_planes,
            image_plane_blurring_image_1d_of_planes=tracer.image_plane_blurring_image_1d_of_planes,
            convolver=convolver_blur, map_to_scaled_array=data_grid_stack.regular.array_2d_from_array_1d)

        assert (blurred_image_of_planes[0] == blurred_plane_image_0).all()
        assert (blurred_image_of_planes[1] == blurred_plane_image_1).all()
        assert (blurred_image_of_planes[2] == blurred_plane_image_2).all()

    def test__blurred_images_of_planes__if_galaxy_has_no_light_profile__replace_with_none(self, mask, convolver_blur):
        data_grid_stack = grids.GridStack.grid_stack_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                           sub_grid_size=1,
                                                                                           psf_shape=(3, 3))

        g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))
        g0_image = galaxy_util.intensities_of_galaxies_from_grid(grid=data_grid_stack.sub, galaxies=[g0])
        g0_blurring_image = galaxy_util.intensities_of_galaxies_from_grid(grid=data_grid_stack.blurring, galaxies=[g0])
        g0_blurred_image_1d = convolver_blur.convolve_image(image_array=g0_image, blurring_array=g0_blurring_image)
        g0_blurred_image = data_grid_stack.regular.scaled_array_2d_from_array_1d(array_1d=g0_blurred_image_1d)

        tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0], source_galaxies=[g0],
                                                     image_plane_grid_stack=data_grid_stack)

        blurred_image_of_planes = util.blurred_image_of_planes_from_1d_images_and_convolver(
            total_planes=tracer.total_planes, image_plane_image_1d_of_planes=tracer.image_plane_image_1d_of_planes,
            image_plane_blurring_image_1d_of_planes=tracer.image_plane_blurring_image_1d_of_planes,
            convolver=convolver_blur, map_to_scaled_array=data_grid_stack.regular.array_2d_from_array_1d)

        assert (blurred_image_of_planes[0] == g0_blurred_image).all()
        assert (blurred_image_of_planes[1] == g0_blurred_image).all()

        tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0], source_galaxies=[g.Galaxy()],
                                                     image_plane_grid_stack=data_grid_stack)

        blurred_image_of_planes = util.blurred_image_of_planes_from_1d_images_and_convolver(
            total_planes=tracer.total_planes, image_plane_image_1d_of_planes=tracer.image_plane_image_1d_of_planes,
            image_plane_blurring_image_1d_of_planes=tracer.image_plane_blurring_image_1d_of_planes,
            convolver=convolver_blur, map_to_scaled_array=data_grid_stack.regular.array_2d_from_array_1d)

        assert (blurred_image_of_planes[0] == g0_blurred_image).all()
        assert blurred_image_of_planes[1] == None

        tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g.Galaxy()], source_galaxies=[g0],
                                                     image_plane_grid_stack=data_grid_stack)

        blurred_image_of_planes = util.blurred_image_of_planes_from_1d_images_and_convolver(
            total_planes=tracer.total_planes, image_plane_image_1d_of_planes=tracer.image_plane_image_1d_of_planes,
            image_plane_blurring_image_1d_of_planes=tracer.image_plane_blurring_image_1d_of_planes,
            convolver=convolver_blur, map_to_scaled_array=data_grid_stack.regular.array_2d_from_array_1d)

        assert blurred_image_of_planes[0] == None
        assert (blurred_image_of_planes[1] == g0_blurred_image).all()

        tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g.Galaxy()], source_galaxies=[g.Galaxy()],
                                                     image_plane_grid_stack=data_grid_stack)

        blurred_image_of_planes = util.blurred_image_of_planes_from_1d_images_and_convolver(
            total_planes=tracer.total_planes, image_plane_image_1d_of_planes=tracer.image_plane_image_1d_of_planes,
            image_plane_blurring_image_1d_of_planes=tracer.image_plane_blurring_image_1d_of_planes,
            convolver=convolver_blur, map_to_scaled_array=data_grid_stack.regular.array_2d_from_array_1d)

        assert blurred_image_of_planes[0] == None
        assert blurred_image_of_planes[1] == None


class TestUnmaskedModelImage:

    def test___3x3_padded_image__no_psf_blurring__produces_padded_image(self):
        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, False, True],
                                        [True, True, True]]), pixel_scale=1.0)

        psf = im.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                      [0.0, 1.0, 0.0],
                                      [0.0, 0.0, 0.0]])), pixel_scale=1.0)

        padded_grid_stack = grids.GridStack.padded_grid_stack_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                                    sub_grid_size=1,
                                                                                                    psf_shape=(3, 3))

        unmasked_blurred_image = padded_grid_stack.unmasked_blurred_image_from_psf_and_unmasked_image(
            psf=psf, unmasked_image_1d=np.ones(25))

        assert (unmasked_blurred_image == np.ones((3, 3))).all()

    def test___3x3_padded_image__simple_psf_blurring__produces_padded_image(self):
        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, False, True],
                                        [True, True, True]]), pixel_scale=1.0)

        psf = im.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                      [0.0, 1.0, 2.0],
                                      [0.0, 0.0, 0.0]])), pixel_scale=1.0)

        padded_grid_stack = grids.GridStack.padded_grid_stack_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                                    sub_grid_size=1,
                                                                                                    psf_shape=(3, 3))

        unmasked_blurred_image = padded_grid_stack.unmasked_blurred_image_from_psf_and_unmasked_image(
            psf=psf, unmasked_image_1d=np.ones(25))

        assert (unmasked_blurred_image == 3.0 * np.ones((3, 3))).all()

    def test___3x3_padded_image__asymmetric_psf_blurring__produces_padded_image(self):
        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, False, True],
                                        [True, True, True]]), pixel_scale=1.0)

        psf = im.PSF(array=(np.array([[0.0, 3.0, 0.0],
                                      [0.0, 1.0, 2.0],
                                      [0.0, 0.0, 0.0]])), pixel_scale=1.0)

        padded_grid_stack = grids.GridStack.padded_grid_stack_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                                    sub_grid_size=1,
                                                                                                    psf_shape=(3, 3))

        unmasked_image_1d = np.zeros(25)
        unmasked_image_1d[12] = 1.0

        unmasked_blurred_image = padded_grid_stack.unmasked_blurred_image_from_psf_and_unmasked_image(
            psf=psf, unmasked_image_1d=unmasked_image_1d)

        assert (unmasked_blurred_image == np.array([[0.0, 3.0, 0.0],
                                                    [0.0, 1.0, 2.0],
                                                    [0.0, 0.0, 0.0]])).all()


class TestUnmaskedModelImageOfPlanes:

    def test___3x3_padded_image__no_psf_blurring(self, galaxy_light):
        psf = im.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                      [0.0, 1.0, 0.0],
                                      [0.0, 0.0, 0.0]])), pixel_scale=1.0)

        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, False, True],
                                        [True, True, True]]), pixel_scale=1.0)

        padded_grid_stack = grids.GridStack.padded_grid_stack_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                                    sub_grid_size=1,
                                                                                                    psf_shape=(3, 3))

        tracer = ray_tracing.TracerImagePlane(lens_galaxies=[galaxy_light], image_plane_grid_stack=padded_grid_stack)

        manual_blurred_image_0 = tracer.image_plane.image_plane_image_1d
        manual_blurred_image_0 = padded_grid_stack.regular.map_to_2d_keep_padded(padded_array_1d=manual_blurred_image_0)
        manual_blurred_image_0 = psf.convolve(array=manual_blurred_image_0)

        unmasked_blurred_image_of_planes = \
            util.unmasked_blurred_image_of_planes_from_padded_grid_stack_and_psf(planes=tracer.planes,
                                                                                 padded_grid_stack=padded_grid_stack,
                                                                                 psf=psf)

        assert (unmasked_blurred_image_of_planes[0] == manual_blurred_image_0[1:4, 1:4]).all()

    def test___x1_galaxy__3x3_padded_image__asymetric_psf_blurring(self, galaxy_light):
        psf = im.PSF(array=(np.array([[0.0, 3.0, 0.0],
                                      [0.0, 1.0, 2.0],
                                      [0.0, 0.0, 0.0]])), pixel_scale=1.0)

        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, False, True],
                                        [True, True, True]]), pixel_scale=1.0)

        padded_grid_stack = grids.GridStack.padded_grid_stack_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                                    sub_grid_size=1,
                                                                                                    psf_shape=(
                                                                                                        3, 3))

        tracer = ray_tracing.TracerImagePlane(lens_galaxies=[galaxy_light], image_plane_grid_stack=padded_grid_stack)

        manual_blurred_image_0 = tracer.image_plane.image_plane_image_1d
        manual_blurred_image_0 = padded_grid_stack.regular.map_to_2d_keep_padded(padded_array_1d=manual_blurred_image_0)
        manual_blurred_image_0 = psf.convolve(array=manual_blurred_image_0)

        unmasked_blurred_image_of_planes = \
            util.unmasked_blurred_image_of_planes_from_padded_grid_stack_and_psf(planes=tracer.planes,
                                                                                 padded_grid_stack=padded_grid_stack,
                                                                                 psf=psf)

        assert (unmasked_blurred_image_of_planes[0] == manual_blurred_image_0[1:4, 1:4]).all()

    def test___x2_galaxies__3x3_padded_image__asymetric_psf_blurring(self):
        psf = im.PSF(array=(np.array([[0.0, 3.0, 0.0],
                                      [0.0, 1.0, 2.0],
                                      [0.0, 0.0, 0.0]])), pixel_scale=1.0)

        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, False, True],
                                        [True, True, True]]), pixel_scale=1.0)

        padded_grid_stack = grids.GridStack.padded_grid_stack_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                                    sub_grid_size=1,
                                                                                                    psf_shape=(3, 3))

        g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.1))
        g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.2))

        tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0, g1], image_plane_grid_stack=padded_grid_stack)

        manual_blurred_image_0 = tracer.image_plane.image_plane_image_1d
        manual_blurred_image_0 = padded_grid_stack.regular.map_to_2d_keep_padded(padded_array_1d=manual_blurred_image_0)
        manual_blurred_image_0 = psf.convolve(array=manual_blurred_image_0)

        unmasked_blurred_image_of_planes = \
            util.unmasked_blurred_image_of_planes_from_padded_grid_stack_and_psf(planes=tracer.planes,
                                                                                 padded_grid_stack=padded_grid_stack,
                                                                                 psf=psf)

        assert unmasked_blurred_image_of_planes[0] == \
               pytest.approx(manual_blurred_image_0[1:4, 1:4], 1.0e-4)

    def test___same_as_above_but_image_and_souce_plane(self):
        psf = im.PSF(array=(np.array([[0.0, 3.0, 0.0],
                                      [0.0, 1.0, 2.0],
                                      [0.0, 0.0, 0.0]])), pixel_scale=1.0)

        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, False, True],
                                        [True, True, True]]), pixel_scale=1.0)

        padded_grid_stack = grids.GridStack.padded_grid_stack_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                                    sub_grid_size=1,
                                                                                                    psf_shape=(3, 3))

        g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.1))
        g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.2))
        g2 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.3))
        g3 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.4))

        tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0, g1], source_galaxies=[g2, g3],
                                                     image_plane_grid_stack=padded_grid_stack)

        manual_blurred_image_0 = tracer.image_plane.image_plane_image_1d
        manual_blurred_image_0 = padded_grid_stack.regular.map_to_2d_keep_padded(padded_array_1d=manual_blurred_image_0)
        manual_blurred_image_0 = psf.convolve(array=manual_blurred_image_0)

        manual_blurred_image_1 = tracer.source_plane.image_plane_image_1d
        manual_blurred_image_1 = padded_grid_stack.regular.map_to_2d_keep_padded(padded_array_1d=manual_blurred_image_1)
        manual_blurred_image_1 = psf.convolve(array=manual_blurred_image_1)

        unmasked_blurred_image_of_planes = \
            util.unmasked_blurred_image_of_planes_from_padded_grid_stack_and_psf(planes=tracer.planes,
                                                                                 padded_grid_stack=padded_grid_stack,
                                                                                 psf=psf)

        assert unmasked_blurred_image_of_planes[0] == pytest.approx(manual_blurred_image_0[1:4, 1:4], 1.0e-4)
        assert unmasked_blurred_image_of_planes[1] == pytest.approx(manual_blurred_image_1[1:4, 1:4], 1.0e-4)

    def test___same_as_above_but_image_and_souce_plane__compare_to_images_of_galaxies(self):
        psf = im.PSF(array=(np.array([[0.0, 3.0, 0.0],
                                      [0.0, 1.0, 2.0],
                                      [0.0, 0.0, 0.0]])), pixel_scale=1.0)

        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, False, True],
                                        [True, True, True]]), pixel_scale=1.0)

        padded_grid_stack = grids.GridStack.padded_grid_stack_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                                    sub_grid_size=1,
                                                                                                    psf_shape=(3, 3))

        g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.1))
        g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.2))
        g2 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.3))
        g3 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.4))

        tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0, g1], source_galaxies=[g2, g3],
                                                     image_plane_grid_stack=padded_grid_stack)

        manual_blurred_image_0 = tracer.image_plane.image_plane_image_1d_of_galaxies[0]
        manual_blurred_image_0 = padded_grid_stack.regular.map_to_2d_keep_padded(padded_array_1d=manual_blurred_image_0)
        manual_blurred_image_0 = psf.convolve(array=manual_blurred_image_0)

        manual_blurred_image_1 = tracer.image_plane.image_plane_image_1d_of_galaxies[1]
        manual_blurred_image_1 = padded_grid_stack.regular.map_to_2d_keep_padded(padded_array_1d=manual_blurred_image_1)
        manual_blurred_image_1 = psf.convolve(array=manual_blurred_image_1)

        manual_blurred_image_2 = tracer.source_plane.image_plane_image_1d_of_galaxies[0]
        manual_blurred_image_2 = padded_grid_stack.regular.map_to_2d_keep_padded(padded_array_1d=manual_blurred_image_2)
        manual_blurred_image_2 = psf.convolve(array=manual_blurred_image_2)

        manual_blurred_image_3 = tracer.source_plane.image_plane_image_1d_of_galaxies[1]
        manual_blurred_image_3 = padded_grid_stack.regular.map_to_2d_keep_padded(padded_array_1d=manual_blurred_image_3)
        manual_blurred_image_3 = psf.convolve(array=manual_blurred_image_3)

        unmasked_blurred_image_of_planes = \
            util.unmasked_blurred_image_of_planes_from_padded_grid_stack_and_psf(planes=tracer.planes,
                                                                                 padded_grid_stack=padded_grid_stack,
                                                                                 psf=psf)

        assert unmasked_blurred_image_of_planes[0] == \
               pytest.approx(manual_blurred_image_0[1:4, 1:4] + manual_blurred_image_1[1:4, 1:4], 1.0e-4)
        assert unmasked_blurred_image_of_planes[1] == \
               pytest.approx(manual_blurred_image_2[1:4, 1:4] + manual_blurred_image_3[1:4, 1:4], 1.0e-4)

    def test__if_plane_has_pixelization__unmasked_image_returns_none(self):
        psf = im.PSF(array=(np.array([[0.0, 3.0, 0.0],
                                      [0.0, 1.0, 2.0],
                                      [0.0, 0.0, 0.0]])), pixel_scale=1.0)

        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, False, True],
                                        [True, True, True]]), pixel_scale=1.0)

        padded_grid_stack = grids.GridStack.padded_grid_stack_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                                    sub_grid_size=1,
                                                                                                    psf_shape=(3, 3))

        g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.1))
        g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.2))
        g_pix = g.Galaxy(pixelization=pix.Rectangular(), regularization=reg.Constant())

        tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0], source_galaxies=[g1, g_pix],
                                                     image_plane_grid_stack=padded_grid_stack)

        unmasked_blurred_image_of_planes = \
            util.unmasked_blurred_image_of_planes_from_padded_grid_stack_and_psf(planes=tracer.planes,
                                                                                 padded_grid_stack=padded_grid_stack,
                                                                                 psf=psf)

        assert type(unmasked_blurred_image_of_planes[0]) == scaled_array.ScaledSquarePixelArray
        assert unmasked_blurred_image_of_planes[1] == None

        tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g_pix], source_galaxies=[g1],
                                                     image_plane_grid_stack=padded_grid_stack)

        unmasked_blurred_image_of_planes = \
            util.unmasked_blurred_image_of_planes_from_padded_grid_stack_and_psf(planes=tracer.planes,
                                                                                 padded_grid_stack=padded_grid_stack,
                                                                                 psf=psf)

        assert unmasked_blurred_image_of_planes[0] == None
        assert type(unmasked_blurred_image_of_planes[1]) == scaled_array.ScaledSquarePixelArray

        tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0, g_pix], source_galaxies=[g_pix],
                                                     image_plane_grid_stack=padded_grid_stack)

        unmasked_blurred_image_of_planes = \
            util.unmasked_blurred_image_of_planes_from_padded_grid_stack_and_psf(planes=tracer.planes,
                                                                                 padded_grid_stack=padded_grid_stack,
                                                                                 psf=psf)

        assert unmasked_blurred_image_of_planes[0] == None
        assert unmasked_blurred_image_of_planes[1] == None


class TestUnmaskedModelImageOfPlanesAndGalaxies:

    def test___3x3_padded_image__no_psf_blurring(self, galaxy_light):
        psf = im.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                      [0.0, 1.0, 0.0],
                                      [0.0, 0.0, 0.0]])), pixel_scale=1.0)

        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, False, True],
                                        [True, True, True]]), pixel_scale=1.0)

        padded_grid_stack = grids.GridStack.padded_grid_stack_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                                    sub_grid_size=1,
                                                                                                    psf_shape=(3, 3))

        tracer = ray_tracing.TracerImagePlane(lens_galaxies=[galaxy_light], image_plane_grid_stack=padded_grid_stack)

        manual_blurred_image_0 = tracer.image_plane.image_plane_image_1d_of_galaxies[0]
        manual_blurred_image_0 = padded_grid_stack.regular.map_to_2d_keep_padded(padded_array_1d=manual_blurred_image_0)
        manual_blurred_image_0 = psf.convolve(array=manual_blurred_image_0)

        unmasked_blurred_image_of_planes_and_galaxies = \
            util.unmasked_blurred_image_of_planes_and_galaxies_from_padded_grid_stack_and_psf(planes=tracer.planes,
                                                                                              padded_grid_stack=padded_grid_stack,
                                                                                              psf=psf)

        assert (unmasked_blurred_image_of_planes_and_galaxies[0][0] == manual_blurred_image_0[1:4, 1:4]).all()

    def test___x1_galaxy__3x3_padded_image__asymetric_psf_blurring(self, galaxy_light):
        psf = im.PSF(array=(np.array([[0.0, 3.0, 0.0],
                                      [0.0, 1.0, 2.0],
                                      [0.0, 0.0, 0.0]])), pixel_scale=1.0)

        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, False, True],
                                        [True, True, True]]), pixel_scale=1.0)

        padded_grid_stack = grids.GridStack.padded_grid_stack_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                                    sub_grid_size=1,
                                                                                                    psf_shape=(
                                                                                                        3, 3))

        tracer = ray_tracing.TracerImagePlane(lens_galaxies=[galaxy_light], image_plane_grid_stack=padded_grid_stack)

        manual_blurred_image_0 = tracer.image_plane.image_plane_image_1d_of_galaxies[0]
        manual_blurred_image_0 = padded_grid_stack.regular.map_to_2d_keep_padded(padded_array_1d=manual_blurred_image_0)
        manual_blurred_image_0 = psf.convolve(array=manual_blurred_image_0)

        unmasked_blurred_image_of_planes_and_galaxies = \
            util.unmasked_blurred_image_of_planes_and_galaxies_from_padded_grid_stack_and_psf(planes=tracer.planes,
                                                                                              padded_grid_stack=padded_grid_stack,
                                                                                              psf=psf)

        assert (unmasked_blurred_image_of_planes_and_galaxies[0][0] == manual_blurred_image_0[1:4, 1:4]).all()

    def test___x2_galaxies__3x3_padded_image__asymetric_psf_blurring(self):
        psf = im.PSF(array=(np.array([[0.0, 3.0, 0.0],
                                      [0.0, 1.0, 2.0],
                                      [0.0, 0.0, 0.0]])), pixel_scale=1.0)

        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, False, True],
                                        [True, True, True]]), pixel_scale=1.0)

        padded_grid_stack = grids.GridStack.padded_grid_stack_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                                    sub_grid_size=1,
                                                                                                    psf_shape=(3, 3))

        g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.1))
        g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.2))

        tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0, g1], image_plane_grid_stack=padded_grid_stack)

        manual_blurred_image_0 = tracer.image_plane.image_plane_image_1d_of_galaxies[0]
        manual_blurred_image_0 = padded_grid_stack.regular.map_to_2d_keep_padded(padded_array_1d=manual_blurred_image_0)
        manual_blurred_image_0 = psf.convolve(array=manual_blurred_image_0)

        manual_blurred_image_1 = tracer.image_plane.image_plane_image_1d_of_galaxies[1]
        manual_blurred_image_1 = padded_grid_stack.regular.map_to_2d_keep_padded(padded_array_1d=manual_blurred_image_1)
        manual_blurred_image_1 = psf.convolve(array=manual_blurred_image_1)

        unmasked_blurred_image_of_planes_and_galaxies = \
            util.unmasked_blurred_image_of_planes_and_galaxies_from_padded_grid_stack_and_psf(planes=tracer.planes,
                                                                                              padded_grid_stack=padded_grid_stack,
                                                                                              psf=psf)

        assert (unmasked_blurred_image_of_planes_and_galaxies[0][0] == manual_blurred_image_0[1:4, 1:4]).all()
        assert (unmasked_blurred_image_of_planes_and_galaxies[0][1] == manual_blurred_image_1[1:4, 1:4]).all()

    def test___same_as_above_but_image_and_souce_plane(self):
        psf = im.PSF(array=(np.array([[0.0, 3.0, 0.0],
                                      [0.0, 1.0, 2.0],
                                      [0.0, 0.0, 0.0]])), pixel_scale=1.0)

        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, False, True],
                                        [True, True, True]]), pixel_scale=1.0)

        padded_grid_stack = grids.GridStack.padded_grid_stack_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                                    sub_grid_size=1,
                                                                                                    psf_shape=(3, 3))

        g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.1))
        g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.2))
        g2 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.3))
        g3 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.4))

        tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0, g1], source_galaxies=[g2, g3],
                                                     image_plane_grid_stack=padded_grid_stack)

        manual_blurred_image_0 = tracer.image_plane.image_plane_image_1d_of_galaxies[0]
        manual_blurred_image_0 = padded_grid_stack.regular.map_to_2d_keep_padded(padded_array_1d=manual_blurred_image_0)
        manual_blurred_image_0 = psf.convolve(array=manual_blurred_image_0)

        manual_blurred_image_1 = tracer.image_plane.image_plane_image_1d_of_galaxies[1]
        manual_blurred_image_1 = padded_grid_stack.regular.map_to_2d_keep_padded(padded_array_1d=manual_blurred_image_1)
        manual_blurred_image_1 = psf.convolve(array=manual_blurred_image_1)

        manual_blurred_image_2 = tracer.source_plane.image_plane_image_1d_of_galaxies[0]
        manual_blurred_image_2 = padded_grid_stack.regular.map_to_2d_keep_padded(padded_array_1d=manual_blurred_image_2)
        manual_blurred_image_2 = psf.convolve(array=manual_blurred_image_2)

        manual_blurred_image_3 = tracer.source_plane.image_plane_image_1d_of_galaxies[1]
        manual_blurred_image_3 = padded_grid_stack.regular.map_to_2d_keep_padded(padded_array_1d=manual_blurred_image_3)
        manual_blurred_image_3 = psf.convolve(array=manual_blurred_image_3)

        unmasked_blurred_image_of_planes_and_galaxies = \
            util.unmasked_blurred_image_of_planes_and_galaxies_from_padded_grid_stack_and_psf(planes=tracer.planes,
                                                                                              padded_grid_stack=padded_grid_stack,
                                                                                              psf=psf)

        assert (unmasked_blurred_image_of_planes_and_galaxies[0][0] == manual_blurred_image_0[1:4, 1:4]).all()
        assert (unmasked_blurred_image_of_planes_and_galaxies[0][1] == manual_blurred_image_1[1:4, 1:4]).all()
        assert (unmasked_blurred_image_of_planes_and_galaxies[1][0] == manual_blurred_image_2[1:4, 1:4]).all()
        assert (unmasked_blurred_image_of_planes_and_galaxies[1][1] == manual_blurred_image_3[1:4, 1:4]).all()

    def test___if_galaxy_has_pixelization__unmasked_image_is_none(self):
        psf = im.PSF(array=(np.array([[0.0, 3.0, 0.0],
                                      [0.0, 1.0, 2.0],
                                      [0.0, 0.0, 0.0]])), pixel_scale=1.0)

        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, False, True],
                                        [True, True, True]]), pixel_scale=1.0)

        padded_grid_stack = grids.GridStack.padded_grid_stack_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                                    sub_grid_size=1,
                                                                                                    psf_shape=(3, 3))

        g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.1),
                      pixelization=pix.Rectangular(), regularization=reg.Constant())
        g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.2))
        g2 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.3))
        g3 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.4))

        tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0, g1], source_galaxies=[g2, g3],
                                                     image_plane_grid_stack=padded_grid_stack)

        unmasked_blurred_image_of_planes_and_galaxies = \
            util.unmasked_blurred_image_of_planes_and_galaxies_from_padded_grid_stack_and_psf(planes=tracer.planes,
                                                                                              padded_grid_stack=padded_grid_stack,
                                                                                              psf=psf)

        assert unmasked_blurred_image_of_planes_and_galaxies[0][0] == None
        assert type(unmasked_blurred_image_of_planes_and_galaxies[0][1]) == scaled_array.ScaledSquarePixelArray
        assert type(unmasked_blurred_image_of_planes_and_galaxies[1][0]) == scaled_array.ScaledSquarePixelArray
        assert type(unmasked_blurred_image_of_planes_and_galaxies[1][1]) == scaled_array.ScaledSquarePixelArray

        g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.1))
        g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.2),
                      pixelization=pix.Rectangular(), regularization=reg.Constant())
        g2 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.3))
        g3 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.4),
                      pixelization=pix.Rectangular(), regularization=reg.Constant())

        tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0, g1], source_galaxies=[g2, g3],
                                                     image_plane_grid_stack=padded_grid_stack)

        unmasked_blurred_image_of_planes_and_galaxies = \
            util.unmasked_blurred_image_of_planes_and_galaxies_from_padded_grid_stack_and_psf(planes=tracer.planes,
                                                                                              padded_grid_stack=padded_grid_stack,
                                                                                              psf=psf)

        assert type(unmasked_blurred_image_of_planes_and_galaxies[0][0]) == scaled_array.ScaledSquarePixelArray
        assert unmasked_blurred_image_of_planes_and_galaxies[0][1] == None
        assert type(unmasked_blurred_image_of_planes_and_galaxies[1][0]) == scaled_array.ScaledSquarePixelArray
        assert unmasked_blurred_image_of_planes_and_galaxies[1][1] == None


class TestContributionsFromHypers:

    def test__x1_hyper_galaxy__model_is_galaxy_image__contributions_all_1(self):
        hyper_galaxies = [MockHyperGalaxy(contribution_factor=1.0, noise_factor=0.0, noise_power=1.0)]

        hyper_model_image_1d = np.array([1.0, 1.0, 1.0, 1.0])

        hyper_galaxy_images_1d = [np.array([1.0, 1.0, 1.0, 1.0])]

        minimum_values = [0.0]

        contribution_maps_1d = util.contribution_maps_1d_from_hyper_images_and_galaxies(
            hyper_model_image_1d=hyper_model_image_1d, hyper_galaxy_images_1d=hyper_galaxy_images_1d,
            hyper_galaxies=hyper_galaxies, hyper_minimum_values=minimum_values)

        assert (contribution_maps_1d[0] == np.array([1.0, 1.0, 1.0, 1.0])).all()

    def test__x1_hyper_galaxy__model_and_galaxy_image_different_contributions_change(self):
        hyper_galaxies = [MockHyperGalaxy(contribution_factor=1.0, noise_factor=0.0, noise_power=1.0)]

        hyper_model_image_1d = np.array([0.5, 1.0, 1.5, 0.5])

        hyper_galaxy_images_1d = [np.array([0.5, 1.0, 1.5, 0.5])]

        minimum_values = [0.6]

        contribution_maps_1d = util.contribution_maps_1d_from_hyper_images_and_galaxies(
            hyper_model_image_1d=hyper_model_image_1d, hyper_galaxy_images_1d=hyper_galaxy_images_1d,
            hyper_galaxies=hyper_galaxies,
            hyper_minimum_values=minimum_values)

        assert (contribution_maps_1d[0] == np.array([0.0, (1.0 / 2.0) / (1.5 / 2.5), 1.0, 0.0])).all()

    def test__x2_hyper_galaxy__model_and_galaxy_image_different_contributions_change(self):
        hyper_galaxies = [MockHyperGalaxy(contribution_factor=0.0, noise_factor=0.0, noise_power=1.0),
                          MockHyperGalaxy(contribution_factor=1.0, noise_factor=0.0, noise_power=1.0)]

        hyper_model_image_1d = np.array([0.5, 1.0, 1.5])

        hyper_galaxy_images_1d = [np.array([0.5, 1.0, 1.5]), np.array([0.5, 1.0, 1.5])]

        minimum_values = [0.5, 0.6]

        contribution_maps_1d = util.contribution_maps_1d_from_hyper_images_and_galaxies(
            hyper_model_image_1d=hyper_model_image_1d, hyper_galaxy_images_1d=hyper_galaxy_images_1d,
            hyper_galaxies=hyper_galaxies,
            hyper_minimum_values=minimum_values)

        assert (contribution_maps_1d[0] == np.array([1.0, 1.0, 1.0])).all()
        assert (contribution_maps_1d[1] == np.array([0.0, (1.0 / 2.0) / (1.5 / 2.5), 1.0])).all()

    def test__x2_hyper_galaxy__same_as_above_use_real_hyper_galaxy(self):
        hyper_galaxies = [g.HyperGalaxy(contribution_factor=0.0, noise_factor=0.0, noise_power=1.0),
                          g.HyperGalaxy(contribution_factor=1.0, noise_factor=0.0, noise_power=1.0)]

        hyper_model_image_1d = np.array([0.5, 1.0, 1.5, 0.5])

        hyper_galaxy_images_1d = [np.array([0.5, 1.0, 1.5, 0.5]), np.array([0.5, 1.0, 1.5, 0.5])]

        minimum_values = [0.5, 0.6]

        contribution_maps_1d = util.contribution_maps_1d_from_hyper_images_and_galaxies(
            hyper_model_image_1d=hyper_model_image_1d, hyper_galaxy_images_1d=hyper_galaxy_images_1d,
            hyper_galaxies=hyper_galaxies,
            hyper_minimum_values=minimum_values)

        assert (contribution_maps_1d[0] == np.array([1.0, 1.0, 1.0, 1.0])).all()
        assert (contribution_maps_1d[1] == np.array([0.0, (1.0 / 2.0) / (1.5 / 2.5), 1.0, 0.0])).all()


class TestScaledNoiseFromContributions:

    def test__x1_hyper_galaxy__noise_factor_is_0__scaled_noise_is_input_noise(self):
        contributions_1d = [np.array([1.0, 1.0, 2.0])]
        hyper_galaxies = [MockHyperGalaxy(contribution_factor=1.0, noise_factor=0.0, noise_power=1.0)]
        noise_map_ = np.array([1.0, 1.0, 1.0])

        scaled_noise_map_ = util.scaled_noise_map_from_hyper_galaxies_and_contribution_maps(
            contribution_maps=contributions_1d, hyper_galaxies=hyper_galaxies, noise_map=noise_map_)

        assert (scaled_noise_map_ == noise_map_).all()

    def test__x1_hyper_galaxy__noise_factor_and_power_are_1__scaled_noise_added_to_input_noise(self):
        contributions_1d = [np.array([1.0, 1.0, 0.5])]
        hyper_galaxies = [MockHyperGalaxy(contribution_factor=1.0, noise_factor=1.0, noise_power=1.0)]
        noise_map_ = np.array([1.0, 1.0, 1.0])

        scaled_noise_map_ = util.scaled_noise_map_from_hyper_galaxies_and_contribution_maps(
            contribution_maps=contributions_1d, hyper_galaxies=hyper_galaxies, noise_map=noise_map_)

        assert (scaled_noise_map_ == np.array([2.0, 2.0, 1.5])).all()

    def test__x1_hyper_galaxy__noise_factor_1_and_power_is_2__scaled_noise_added_to_input_noise(self):
        contributions_1d = [np.array([1.0, 1.0, 0.5])]
        hyper_galaxies = [MockHyperGalaxy(contribution_factor=1.0, noise_factor=1.0, noise_power=2.0)]
        noise_map_ = np.array([1.0, 1.0, 1.0])

        scaled_noise_map_ = util.scaled_noise_map_from_hyper_galaxies_and_contribution_maps(
            contribution_maps=contributions_1d, hyper_galaxies=hyper_galaxies, noise_map=noise_map_)

        assert (scaled_noise_map_ == np.array([2.0, 2.0, 1.25])).all()

    def test__x2_hyper_galaxy__noise_factor_1_and_power_is_2__scaled_noise_added_to_input_noise(self):
        contributions_1d = [np.array([1.0, 1.0, 0.5]), np.array([0.25, 0.25, 0.25])]
        hyper_galaxies = [MockHyperGalaxy(contribution_factor=1.0, noise_factor=1.0, noise_power=2.0),
                          MockHyperGalaxy(contribution_factor=1.0, noise_factor=2.0, noise_power=1.0)]
        noise_map_ = np.array([1.0, 1.0, 1.0])

        scaled_noise_map_ = util.scaled_noise_map_from_hyper_galaxies_and_contribution_maps(
            contribution_maps=contributions_1d, hyper_galaxies=hyper_galaxies, noise_map=noise_map_)

        assert (scaled_noise_map_ == np.array([2.5, 2.5, 1.75])).all()

    def test__x2_hyper_galaxy__same_as_above_but_use_real_hyper_galaxy(self):
        contributions_1d = [np.array([1.0, 1.0, 0.5]), np.array([0.25, 0.25, 0.25])]
        hyper_galaxies = [g.HyperGalaxy(contribution_factor=1.0, noise_factor=1.0, noise_power=2.0),
                          g.HyperGalaxy(contribution_factor=1.0, noise_factor=2.0, noise_power=1.0)]
        noise_map_ = np.array([1.0, 1.0, 1.0])

        scaled_noise_map_ = util.scaled_noise_map_from_hyper_galaxies_and_contribution_maps(
            contribution_maps=contributions_1d, hyper_galaxies=hyper_galaxies, noise_map=noise_map_)

        assert (scaled_noise_map_ == np.array([2.5, 2.5, 1.75])).all()

    # def test__same_as_above_but_using_fiting_hyper_image(self, image, masks):
    #
    #     fitting_hyper_image = fit_data.FitDataHyper(datas=image, masks=masks, hyper_model_image=None,
    #                                                 hyper_galaxy_images=None, hyper_minimum_values=None)
    #
    #     contribution_maps_1d = [np.array([1.0, 1.0, 0.5]), np.array([0.25, 0.25, 0.25])]
    #     hyper_galaxies = [g.HyperGalaxy(contribution_factor=1.0, noise_factor=1.0, noise_power=2.0),
    #                       g.HyperGalaxy(contribution_factor=1.0, noise_factor=2.0, noise_power=1.0)]
    #     scaled_noises = lens_fit_util.scaled_noise_maps_from_fitting_hyper_images_contributions_and_hyper_galaxies(
    #         fitting_hyper_images=[fitting_hyper_image], contributions_1d=[contribution_maps_1d], hyper_galaxies=hyper_galaxies)
    #
    #     assert (scaled_noises[0] == np.array([2.5, 2.5, 1.75])).all()
    #
    # def test__same_as_above_but_using_x2_fiting_hyper_image(self, image, masks):
    #
    #     fitting_hyper_image_0 = fit_data.FitDataHyper(datas=image, masks=masks, hyper_model_image=None,
    #                                                   hyper_galaxy_images=None, hyper_minimum_values=None)
    #     fitting_hyper_image_1 = fit_data.FitDataHyper(datas=image, masks=masks, hyper_model_image=None,
    #                                                   hyper_galaxy_images=None, hyper_minimum_values=None)
    #
    #     contributions_0 = [np.array([1.0, 1.0, 0.5]), np.array([0.25, 0.25, 0.25])]
    #     contributions_1 = [np.array([1.0, 1.0, 0.5]), np.array([0.25, 0.25, 0.25])]
    #     hyper_galaxies = [g.HyperGalaxy(contribution_factor=1.0, noise_factor=1.0, noise_power=2.0),
    #                       g.HyperGalaxy(contribution_factor=1.0, noise_factor=2.0, noise_power=1.0)]
    #     scaled_noises = lens_fit_util.scaled_noise_maps_from_fitting_hyper_images_contributions_and_hyper_galaxies(
    #         fitting_hyper_images=[fitting_hyper_image_0, fitting_hyper_image_1],
    #         contributions_1d=[contributions_0, contributions_1], hyper_galaxies=hyper_galaxies)
    #
    #     assert (scaled_noises[0] == np.array([2.5, 2.5, 1.75])).all()
    #     assert (scaled_noises[1] == np.array([2.5, 2.5, 1.75])).all()



import time

from autolens.data.array import grids
from autolens.model.profiles import mass_profiles as mp
from autolens.lens import lens_data as ld
from autolens.data.array import mask as msk

from autolens.data.array.plotters import array_plotters

from test.simulation import simulation_util

import numpy as np
import matplotlib.pyplot as plt

# Although we could test the deflection angles without using an image (e.g. by just making a grid), we have chosen to
# set this test up using an image and mask. This gives run-time numbers that can be easily related to an actual lens
# analysis

sub_grid_size = 2
inner_radius_arcsec = 0.0
outer_radius_arcsec = 4.0

print('sub grid size = ' + str(sub_grid_size))
print('annular inner mask radius = ' + str(inner_radius_arcsec) + '\n')
print('annular outer mask radius = ' + str(outer_radius_arcsec) + '\n')

for data_resolution in ['HST_Up']:

    print()

    ccd_data = simulation_util.load_test_ccd_data(data_type='no_lens_source_smooth', data_resolution=data_resolution,
                                                  psf_shape=(3, 3))
    mask = msk.Mask.circular_annular(shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale,
                                     inner_radius_arcsec=inner_radius_arcsec, outer_radius_arcsec=outer_radius_arcsec)
    lens_data = ld.LensData(ccd_data=ccd_data, mask=mask, sub_grid_size=sub_grid_size)

    print('Deflection angle run times for image type ' + data_resolution + '\n')
    print('Number of points = ' + str(lens_data.grid_stack.regular.shape[0]) + '\n')

    interpolator = grids.Interpolator.from_mask_grid_and_interp_pixel_scales(mask=lens_data.mask,
                                                                             grid=lens_data.grid_stack.sub,
                                                                             interp_pixel_scale=0.1)

    print('Number of interpolation points = ' + str(interpolator.interp_grid.shape[0]) + '\n')

    ### EllipticalIsothermal ###

    mass_profile = mp.EllipticalCoredIsothermal(centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, einstein_radius=0.5,
                                                core_radius=0.3)

    interp_deflections = mass_profile.deflections_from_grid(grid=interpolator.interp_grid)
    deflections = np.zeros((lens_data.grid_stack.sub.shape[0], 2))
    deflections[:,0] = interpolator.interpolated_values_from_values(values=interp_deflections[:,0])
    deflections[:,1] = interpolator.interpolated_values_from_values(values=interp_deflections[:,1])

    true_deflections = mass_profile.deflections_from_grid(grid=lens_data.grid_stack.sub)

    difference_y = deflections[:,0] - true_deflections[:, 0]
    difference_x = deflections[:,1] - true_deflections[:, 1]

    print("interpolation y error: ", np.mean(difference_y))
    print("interpolation y uncertainty: ", np.std(difference_y))
    print("interpolation y max error: ", np.max(difference_y))
    print("interpolation x error: ", np.mean(difference_x))
    print("interpolation x uncertainty: ", np.std(difference_x))
    print("interpolation x max error: ", np.max(difference_x))

    difference_y_2d = lens_data.grid_stack.sub.scaled_array_2d_with_sub_dimensions_from_sub_array_1d(
        sub_array_1d=difference_y)
    difference_x_2d = lens_data.grid_stack.sub.scaled_array_2d_with_sub_dimensions_from_sub_array_1d(
        sub_array_1d=difference_x)

    array_plotters.plot_array(array=difference_y_2d)


from autolens.data.array import mask as msk
from autolens.model.profiles import mass_profiles as mp
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg
from autolens.model.galaxy import galaxy as g
from autolens.lens import ray_tracing
from autolens.lens import lens_fit
from autolens.lens import lens_data as ld
from autolens.data.plotters import ccd_plotters
from autolens.lens.plotters import lens_fit_plotters
from test.simulation import simulation_util

ccd_data = simulation_util.load_test_ccd_data(data_type='no_lens_light_and_source_smooth_offset_centre',
                                              data_resolution='LSST')

def mask_and_fit_with_offset_centre(centre):

    mask = msk.Mask.elliptical(shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, major_axis_radius_arcsec=3.0,
                               axis_ratio=1.0, phi=0.0, centre=centre)

    # The lines of code below do everything we're used to, that is, setup an image and its grid stack, mask it, trace it
    # via a tracer, setup the rectangular mapper, etc.
    lens_galaxy = g.Galaxy(mass=mp.EllipticalIsothermal(centre=(1.0, 1.0), einstein_radius=1.6,
                                                            axis_ratio=0.7, phi=45.0))
    source_galaxy = g.Galaxy(pixelization=pix.AdaptiveMagnification(shape=(20, 20)),
                             regularization=reg.Constant(coefficients=(1.0,)))

    lens_data = ld.LensData(ccd_data=ccd_data, mask=mask)

    tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                                 image_plane_grid_stack=lens_data.grid_stack)
    fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)

    return mask, fit

mask, fit = mask_and_fit_with_offset_centre(centre=(1.0, 1.0))

lens_fit_plotters.plot_fit_subplot(fit=fit, should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True,
                                   positions=[[[2.2, 2.2], [-0.2, -0.2], [-0.2, 2.2], [2.2, -0.2]]],
                                   should_plot_image_plane_pix=True)

mask, fit = mask_and_fit_with_offset_centre(centre=(1.05, 1.05))


lens_fit_plotters.plot_fit_subplot(fit=fit, should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True,
                                   positions=[[[2.2, 2.2], [-0.2, -0.2], [-0.2, 2.2], [2.2, -0.2]]],
                                   should_plot_image_plane_pix=True)

mask, fit = mask_and_fit_with_offset_centre(centre=(1.1, 1.1))

lens_fit_plotters.plot_fit_subplot(fit=fit, should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True,
                                   positions=[[[2.2, 2.2], [-0.2, -0.2], [-0.2, 2.2], [2.2, -0.2]]],
                                   should_plot_image_plane_pix=True)

mask, fit = mask_and_fit_with_offset_centre(centre=(0.95, 0.95))

lens_fit_plotters.plot_fit_subplot(fit=fit, should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True,
                                   positions=[[[2.2, 2.2], [-0.2, -0.2], [-0.2, 2.2], [2.2, -0.2]]],
                                   should_plot_image_plane_pix=True)

mask, fit = mask_and_fit_with_offset_centre(centre=(0.9, 0.9))

lens_fit_plotters.plot_fit_subplot(fit=fit, should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True,
                                   positions=[[[2.2, 2.2], [-0.2, -0.2], [-0.2, 2.2], [2.2, -0.2]]],
                                   should_plot_image_plane_pix=True)


from autolens.data.array import grids
from autolens.model.profiles import mass_profiles as mp
from autolens.model.profiles.plotters import profile_plotters

import numpy as np

regular = grids.RegularGrid.from_shape_and_pixel_scale(shape=(100, 100), pixel_scale=0.05)

axis_ratio = 1.0

sie_02 = mp.EllipticalIsothermal(centre=(0.0, 0.0), einstein_radius=0.2, axis_ratio=1.0)
sie_04 = mp.EllipticalIsothermal(centre=(0.0, 0.0), einstein_radius=0.4, axis_ratio=1.0)
sie_06 = mp.EllipticalIsothermal(centre=(0.0, 0.0), einstein_radius=0.6, axis_ratio=1.0)
sie_08 = mp.EllipticalIsothermal(centre=(0.0, 0.0), einstein_radius=0.8, axis_ratio=1.0)
sie_10 = mp.EllipticalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0, axis_ratio=1.0)

print(sie_02.mass_within_circle_in_angular_units(radius=sie_02.einstein_radius))
print(sie_04.mass_within_circle_in_angular_units(radius=sie_04.einstein_radius))
print(sie_06.mass_within_circle_in_angular_units(radius=sie_06.einstein_radius))
print(sie_08.mass_within_circle_in_angular_units(radius=sie_08.einstein_radius))
print(sie_10.mass_within_circle_in_angular_units(radius=sie_10.einstein_radius))
print()

print(sie_02.mass_within_circle_in_angular_units(radius=sie_02.einstein_radius) / 0.2)
print(sie_04.mass_within_circle_in_angular_units(radius=sie_04.einstein_radius) / 0.4)

sie_02_einstein_mass = sie_02.mass_within_circle_in_angular_units(sie_02.einstein_radius)
sie_04_einstein_mass = sie_04.mass_within_circle_in_angular_units(sie_04.einstein_radius)
sie_06_einstein_mass = sie_06.mass_within_circle_in_angular_units(sie_06.einstein_radius)
sie_08_einstein_mass = sie_08.mass_within_circle_in_angular_units(sie_08.einstein_radius)
sie_10_einstein_mass = sie_10.mass_within_circle_in_angular_units(sie_10.einstein_radius)

print()
print(sie_02_einstein_mass / 0.2)
print(sie_04_einstein_mass / 0.4)
print(sie_06_einstein_mass / 0.6)
print(sie_08_einstein_mass / 0.8)
print(sie_10_einstein_mass / 1.0)

nfw = mp.SphericalNFW(centre=(0.0, 0.0), kappa_s=0.2, scale_radius=8.0)

# profile_plotters.plot_surface_density(mass_profile=sis, grid=regular)

mass = sis.mass_within_circle_in_angular_units(radius=1.0)

print(sis.radius_where_average_convergence_in_circle_is_one)

radius = 0.8116039549729345

mass = nfw.mass_within_circle_in_angular_units(radius=radius)
area = np.pi * radius ** 2.0

print(mass / area)

print(nfw.radius_where_average_convergence_in_circle_is_one)


import numpy as np
from autolens import decorator_util

@decorator_util.jit()
def mapping_matrix_from_sub_to_pix(sub_to_pix, pixels, regular_pixels, sub_to_regular, sub_grid_fraction):
    """Computes the mapping matrix, by iterating over the known mappings between the sub-grid and pixelization.

    Parameters
    -----------
    sub_to_pix : ndarray
        The mappings between the observed regular's sub-pixels and pixelization's pixels.
    pixels : int
        The number of pixels in the pixelization.
    regular_pixels : int
        The number of datas pixels in the observed datas and thus on the regular grid.
    sub_to_regular : ndarray
        The mappings between the observed regular's sub-pixels and observed regular's pixels.
    sub_grid_fraction : float
        The fractional area each sub-pixel takes up in an regular-pixel.
    """

    mapping_matrix = np.zeros((regular_pixels, pixels))

    for sub_index in range(sub_to_regular.shape[0]):
        mapping_matrix[sub_to_regular[sub_index], sub_to_pix[sub_index]] += sub_grid_fraction

    return mapping_matrix

@decorator_util.jit()
def voronoi_regular_to_pix_from_grids_and_geometry(regular_grid, regular_to_nearest_pix, pixel_centres,
                                                   pixel_neighbors, pixel_neighbors_size):
    """ Compute the mappings between a set of regular-grid pixels and pixelization pixels, using information on \
    how regular pixels map to their closest pixelization pixel on the image-plane pix-grid and the pixelization's \
    pixel centres.

    To determine the complete set of regular-pixel to pixelization pixel mappings, we must pair every regular-pixel to \
    its nearest pixel. Using a full nearest neighbor search to do this is slow, thus the pixel neighbors (derived via \
    the Voronoi grid) are used to localize each nearest neighbor search via a graph search.

    Parameters
    ----------
    regular_grid : RegularGrid
        The grid of (y,x) arc-second coordinates at the centre of every unmasked pixel, which has been traced to \
        to an irregular grid via lens.
    regular_to_nearest_pix : ndarray
        A 1D array that maps every regular-grid pixel to its nearest pix-grid pixel (as determined on the unlensed \
        2D array).
    pixel_centres : ndarray
        The (y,x) centre of every Voronoi pixel in arc-seconds.
    pixel_neighbors : ndarray
        An array of length (voronoi_pixels) which provides the index of all neighbors of every pixel in \
        the Voronoi grid (entries of -1 correspond to no neighbor).
    pixel_neighbors_size : ndarray
        An array of length (voronoi_pixels) which gives the number of neighbors of every pixel in the \
        Voronoi grid.
     """

    regular_to_pix = np.zeros((regular_grid.shape[0]))

    for regular_index in range(regular_grid.shape[0]):

        nearest_pix_pixel_index = regular_to_nearest_pix[regular_index]

        while True:

            nearest_pix_pixel_center = pixel_centres[nearest_pix_pixel_index]

            sub_to_nearest_pix_distance = (regular_grid[regular_index, 0] - nearest_pix_pixel_center[0]) ** 2 + \
                                          (regular_grid[regular_index, 1] - nearest_pix_pixel_center[1]) ** 2

            closest_separation_from_pix_neighbor = 1.0e8

            for neighbor_index in range(pixel_neighbors_size[nearest_pix_pixel_index]):

                neighbor = pixel_neighbors[nearest_pix_pixel_index, neighbor_index]

                separation_from_neighbor = (regular_grid[regular_index, 0] - pixel_centres[neighbor, 0]) ** 2 + \
                                           (regular_grid[regular_index, 1] - pixel_centres[neighbor, 1]) ** 2

                if separation_from_neighbor < closest_separation_from_pix_neighbor:

                    closest_separation_from_pix_neighbor = separation_from_neighbor
                    closest_neighbor_index = neighbor_index

            neighboring_pix_pixel_index = pixel_neighbors[nearest_pix_pixel_index, closest_neighbor_index]
            sub_to_neighboring_pix_distance = closest_separation_from_pix_neighbor

            if sub_to_nearest_pix_distance <= sub_to_neighboring_pix_distance:
                regular_to_pix[regular_index] = nearest_pix_pixel_index
                break
            else:
                nearest_pix_pixel_index = neighboring_pix_pixel_index

    return regular_to_pix

@decorator_util.jit()
def voronoi_sub_to_pix_from_grids_and_geometry(sub_grid, regular_to_nearest_pix, sub_to_regular, pixel_centres,
                                               pixel_neighbors, pixel_neighbors_size):
    """ Compute the mappings between a set of sub-grid pixels and pixelization pixels, using information on \
    how the regular pixels hosting each sub-pixel map to their closest pixelization pixel on the image-plane pix-grid \
    and the pixelization's pixel centres.

    To determine the complete set of sub-pixel to pixelization pixel mappings, we must pair every sub-pixel to \
    its nearest pixel. Using a full nearest neighbor search to do this is slow, thus the pixel neighbors (derived via \
    the Voronoi grid) are used to localize each nearest neighbor search by using a graph search.

    Parameters
    ----------
    regular_grid : RegularGrid
        The grid of (y,x) arc-second coordinates at the centre of every unmasked pixel, which has been traced to \
        to an irregular grid via lens.
    regular_to_nearest_pix : ndarray
        A 1D array that maps every regular-grid pixel to its nearest pix-grid pixel (as determined on the unlensed \
        2D array).
    pixel_centres : (float, float)
        The (y,x) centre of every Voronoi pixel in arc-seconds.
    pixel_neighbors : ndarray
        An array of length (voronoi_pixels) which provides the index of all neighbors of every pixel in \
        the Voronoi grid (entries of -1 correspond to no neighbor).
    pixel_neighbors_size : ndarray
        An array of length (voronoi_pixels) which gives the number of neighbors of every pixel in the \
        Voronoi grid.
     """

    sub_to_pix = np.zeros((sub_grid.shape[0]))

    for sub_index in range(sub_grid.shape[0]):

        nearest_pix_pixel_index = regular_to_nearest_pix[sub_to_regular[sub_index]]

        while True:

            nearest_pix_pixel_center = pixel_centres[nearest_pix_pixel_index]

            sub_to_nearest_pix_distance = (sub_grid[sub_index, 0] - nearest_pix_pixel_center[0]) ** 2 + \
                                          (sub_grid[sub_index, 1] - nearest_pix_pixel_center[1]) ** 2

            closest_separation_from_pix_to_neighbor = 1.0e8

            for neighbor_index in range(pixel_neighbors_size[nearest_pix_pixel_index]):

                neighbor = pixel_neighbors[nearest_pix_pixel_index, neighbor_index]

                separation_from_neighbor = (sub_grid[sub_index, 0] - pixel_centres[neighbor, 0]) ** 2 + \
                                           (sub_grid[sub_index, 1] - pixel_centres[neighbor, 1]) ** 2

                if separation_from_neighbor < closest_separation_from_pix_to_neighbor:
                    closest_separation_from_pix_to_neighbor = separation_from_neighbor
                    closest_neighbor_index = neighbor_index

            neighboring_pix_pixel_index = pixel_neighbors[nearest_pix_pixel_index, closest_neighbor_index]
            sub_to_neighboring_pix_distance = closest_separation_from_pix_to_neighbor

            if sub_to_nearest_pix_distance <= sub_to_neighboring_pix_distance:
                sub_to_pix[sub_index] = nearest_pix_pixel_index
                break
            else:
                nearest_pix_pixel_index = neighboring_pix_pixel_index

    return sub_to_pix


def phase_tag_from_phase_settings(sub_grid_size, bin_up_factor, image_psf_shape, inversion_psf_shape,
                                  positions_threshold, inner_mask_radii, interp_pixel_scale):
    
    sub_grid_size_tag = sub_grid_size_tag_from_sub_grid_size(sub_grid_size=sub_grid_size)
    bin_up_factor_tag = bin_up_factor_tag_from_bin_up_factor(bin_up_factor=bin_up_factor)
    image_psf_shape_tag = image_psf_shape_tag_from_image_psf_shape(image_psf_shape=image_psf_shape)
    inversion_psf_shape_tag = inversion_psf_shape_tag_from_inversion_psf_shape(inversion_psf_shape=inversion_psf_shape)
    positions_threshold_tag = positions_threshold_tag_from_positions_threshold(positions_threshold=positions_threshold)
    inner_mask_radii_tag = inner_mask_radii_tag_from_inner_circular_mask_radii(inner_mask_radii=inner_mask_radii)
    interp_pixel_scale_tag = interp_pixel_scale_tag_from_interp_pixel_scale(interp_pixel_scale=interp_pixel_scale)

    return sub_grid_size_tag + bin_up_factor_tag + image_psf_shape_tag + inversion_psf_shape_tag + \
           positions_threshold_tag + inner_mask_radii_tag + interp_pixel_scale_tag

def positions_threshold_tag_from_positions_threshold(positions_threshold):
    """Generate a bin up tag, to customize phase names based on the bin up factor used in a pipeline. This changes
    the phase name 'phase_name' as follows:

    positions_threshold = 1 -> phase_name
    positions_threshold = 2 -> phase_name_positions_threshold_2
    positions_threshold = 2 -> phase_name_positions_threshold_2
    """
    if positions_threshold == None:
        return ''
    else:
        return '_pos_{0:.2f}'.format(positions_threshold)


def sub_grid_size_tag_from_sub_grid_size(sub_grid_size):
    """Generate a bin up tag, to customize phase names based on the bin up factor used in a pipeline. This changes
    the phase name 'phase_name' as follows:

    sub_grid_size = None -> phase_name
    sub_grid_size = 1 -> phase_name_sub_grid_size_2
    sub_grid_size = 4 -> phase_name_sub_grid_size_4
    """
    return '_sub_' + str(sub_grid_size)


def inner_mask_radii_tag_from_inner_circular_mask_radii(inner_mask_radii):
    """Generate a bin up tag, to customize phase names based on the bin up factor used in a pipeline. This changes
    the phase name 'phase_name' as follows:

    inner_circular_mask_radii = 1 -> phase_name
    inner_circular_mask_radii = 2 -> phase_name_inner_circular_mask_radii_2
    inner_circular_mask_radii = 2 -> phase_name_inner_circular_mask_radii_2
    """
    if inner_mask_radii == None:
        return ''
    else:
        return '_inner_mask_{0:.2f}'.format(inner_mask_radii)


def image_psf_shape_tag_from_image_psf_shape(image_psf_shape):
    """Generate a bin up tag, to customize phase names based on the bin up factor used in a pipeline. This changes
    the phase name 'phase_name' as follows:

    image_psf_shape = 1 -> phase_name
    image_psf_shape = 2 -> phase_name_image_psf_shape_2
    image_psf_shape = 2 -> phase_name_image_psf_shape_2
    """
    if image_psf_shape is None:
        return ''
    else:
        y = str(image_psf_shape[0])
        x = str(image_psf_shape[1])
        return ('_image_psf_' + y + 'x' + x)


def inversion_psf_shape_tag_from_inversion_psf_shape(inversion_psf_shape):
    """Generate a bin up tag, to customize phase names based on the bin up factor used in a pipeline. This changes
    the phase name 'phase_name' as follows:

    inversion_psf_shape = 1 -> phase_name
    inversion_psf_shape = 2 -> phase_name_inversion_psf_shape_2
    inversion_psf_shape = 2 -> phase_name_inversion_psf_shape_2
    """
    if inversion_psf_shape is None:
        return ''
    else:
        y = str(inversion_psf_shape[0])
        x = str(inversion_psf_shape[1])
        return ('_inversion_psf_' + y + 'x' + x)


def bin_up_factor_tag_from_bin_up_factor(bin_up_factor):
    """Generate a bin up tag, to customize phase names based on the bin up factor used in a pipeline. This changes
    the phase name 'phase_name' as follows:

    bin_up_factor = 1 -> phase_name
    bin_up_factor = 2 -> phase_name_bin_up_factor_2
    bin_up_factor = 2 -> phase_name_bin_up_factor_2
    """
    if bin_up_factor == 1 or bin_up_factor is None:
        return ''
    else:
        return '_bin_up_' + str(bin_up_factor)


def interp_pixel_scale_tag_from_interp_pixel_scale(interp_pixel_scale):
    """Generate a bin up tag, to customize phase names based on the bin up factor used in a pipeline. This changes
    the phase name 'phase_name' as follows:

    interp_pixel_scale = 1 -> phase_name
    interp_pixel_scale = 2 -> phase_name_interp_pixel_scale_2
    interp_pixel_scale = 2 -> phase_name_interp_pixel_scale_2
    """
    if interp_pixel_scale is None:
        return ''
    else:
        return '_interp_{0:.3f}'.format(interp_pixel_scale)


import time

from autolens.data.array import mask as msk
from autolens.model.galaxy import galaxy as g
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.lens import lens_data as ld
from autolens.lens import ray_tracing
from autolens.lens import lens_fit
from autolens.lens.util import lens_fit_util

from test.simulation import simulation_util

repeats = 10

print('Number of repeats = ' + str(repeats))
print()

sub_grid_size = 4
radius_arcsec = 3.0
psf_shape = (21,21)

print('sub grid size = ' + str(sub_grid_size))
print('circular mask radius = ' + str(radius_arcsec) + '\n')
print('psf shape = ' + str(psf_shape) + '\n')

lens_galaxy = g.Galaxy(light=lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.9, phi=45.0,
                                                intensity=0.5, effective_radius=0.8, sersic_index=4.0),
                       mass=mp.EllipticalIsothermal(centre=(0.0, 0.0), einstein_radius=1.6,
                                                    axis_ratio=0.7, phi=45.0))

source_galaxy = g.Galaxy(light=lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.8, phi=60.0,
                                                   intensity=0.4, effective_radius=0.5, sersic_index=1.0))

for data_resolution in ['LSST', 'Euclid', 'HST', 'HST_Up', 'AO']:

    ccd_data = simulation_util.load_test_ccd_data(data_type='lens_and_source_smooth', data_resolution=data_resolution,
                                                  psf_shape=psf_shape)
    mask = msk.Mask.circular(shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=radius_arcsec)
    lens_data = ld.LensData(ccd_data=ccd_data, mask=mask, sub_grid_size=sub_grid_size)

    print('Light profile fit run times for image type ' + data_resolution + '\n')
    print('Number of points = ' + str(lens_data.grid_stack.sub.shape[0]) + '\n')

    start_overall = time.time()

    start = time.time()
    for i in range(repeats):
        tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                                 image_plane_grid_stack=lens_data.grid_stack)
    diff = time.time() - start
    print("Time to Setup Tracer = {}".format(diff/repeats))

    start = time.time()
    for i in range(repeats):
        blurred_profile_image_1d = lens_fit_util.blurred_image_1d_from_1d_unblurred_and_blurring_images(
            unblurred_image_1d=tracer.image_plane_image_1d, blurring_image_1d=tracer.image_plane_blurring_image_1d,
            convolver=lens_data.convolver_image)
        blurred_profile_image = lens_data.map_to_scaled_array(array_1d=blurred_profile_image_1d)
    diff = time.time() - start
    print("Time to perform PSF convolution = {}".format(diff/repeats))

    start = time.time()
    for i in range(repeats):
        lens_fit.LensDataFit(image=lens_data.image_1d, noise_map=lens_data.noise_map_1d, mask=lens_data.mask_1d,
                               model_image=blurred_profile_image_1d)
    diff = time.time() - start
    print("Time to perform fit (1D) = {}".format(diff/repeats))

    start = time.time()
    for i in range(repeats):
        lens_fit.LensDataFit(image=lens_data.image, noise_map=lens_data.noise_map, mask=lens_data.mask,
                         model_image=blurred_profile_image)
    diff = time.time() - start
    print("Time to perform fit (2D) = {}".format(diff/repeats))

    start = time.time()
    for i in range(repeats):
        lens_fit.LensProfileFit(lens_data=lens_data, tracer=tracer)
    diff = time.time() - start
    print("Time to perform complete fit = {}".format(diff/repeats))

    print()


import numpy as np
from autolens import decorator_util

@decorator_util.jit()
def rectangular_neighbors_from_shape(shape):
    """Compute the neighbors of every pixel as a list of the pixel index's each pixel shares a vertex with.

    The uniformity of the rectangular grid's geometry is used to compute this.
    """


    pixels = shape[0]*shape[1]

    pixel_neighbors = -1 * np.ones(shape=(pixels, 4))
    pixel_neighbors_size = np.zeros(pixels)

    pixel_neighbors, pixel_neighbors_size = compute_corner_neighbors(pixel_neighbors, pixel_neighbors_size,
                                                                     shape, pixels)
    pixel_neighbors, pixel_neighbors_size = compute_top_edge_neighbors(pixel_neighbors, pixel_neighbors_size,
                                                                       shape, pixels)
    pixel_neighbors, pixel_neighbors_size = compute_left_edge_neighbors(pixel_neighbors, pixel_neighbors_size,
                                                                        shape, pixels)
    pixel_neighbors, pixel_neighbors_size = compute_right_edge_neighbors(pixel_neighbors, pixel_neighbors_size,
                                                                         shape, pixels)
    pixel_neighbors, pixel_neighbors_size = compute_bottom_edge_neighbors(pixel_neighbors, pixel_neighbors_size,
                                                                          shape, pixels)
    pixel_neighbors, pixel_neighbors_size = compute_central_neighbors(pixel_neighbors, pixel_neighbors_size,
                                                                      shape, pixels)

    return pixel_neighbors, pixel_neighbors_size

@decorator_util.jit()
def compute_corner_neighbors(pixel_neighbors, pixel_neighbors_size, shape, pixels):

    pixel_neighbors[0, 0:2] = np.array([1, shape[1]])
    pixel_neighbors_size[0] = 2

    pixel_neighbors[shape[1] - 1, 0:2] = np.array([shape[1] - 2, shape[1] + shape[1] - 1])
    pixel_neighbors_size[shape[1] - 1] = 2

    pixel_neighbors[pixels - shape[1], 0:2] = np.array([pixels - shape[1] * 2,
                                                        pixels - shape[1] + 1])
    pixel_neighbors_size[pixels - shape[1]] = 2

    pixel_neighbors[pixels - 1, 0:2] = np.array([pixels - shape[1] - 1, pixels - 2])
    pixel_neighbors_size[pixels - 1] = 2

    return pixel_neighbors, pixel_neighbors_size

@decorator_util.jit()
def compute_top_edge_neighbors(pixel_neighbors, pixel_neighbors_size, shape, pixels):

    for pix in range(1, shape[1] - 1):
        pixel_index = pix
        pixel_neighbors[pixel_index, 0:3] = np.array([pixel_index - 1, pixel_index + 1,
                                                      pixel_index + shape[1]])
        pixel_neighbors_size[pixel_index] = 3

    return pixel_neighbors, pixel_neighbors_size

@decorator_util.jit()
def compute_left_edge_neighbors(pixel_neighbors, pixel_neighbors_size, shape, pixels):

    for pix in range(1, shape[0] - 1):
        pixel_index = pix * shape[1]
        pixel_neighbors[pixel_index, 0:3] = np.array([pixel_index - shape[1], pixel_index + 1,
                                                      pixel_index + shape[1]])
        pixel_neighbors_size[pixel_index] = 3

    return pixel_neighbors, pixel_neighbors_size

@decorator_util.jit()
def compute_right_edge_neighbors(pixel_neighbors, pixel_neighbors_size, shape, pixels):

    for pix in range(1, shape[0] - 1):
        pixel_index = pix * shape[1] + shape[1] - 1
        pixel_neighbors[pixel_index, 0:3] = np.array([pixel_index - shape[1], pixel_index - 1,
                                                      pixel_index + shape[1]])
        pixel_neighbors_size[pixel_index] = 3

    return pixel_neighbors, pixel_neighbors_size

@decorator_util.jit()
def compute_bottom_edge_neighbors(pixel_neighbors, pixel_neighbors_size, shape, pixels):

    for pix in range(1, shape[1] - 1):
        pixel_index = pixels - pix - 1
        pixel_neighbors[pixel_index, 0:3] = np.array([pixel_index - shape[1], pixel_index - 1,
                                                      pixel_index + 1])
        pixel_neighbors_size[pixel_index] = 3

    return pixel_neighbors, pixel_neighbors_size

@decorator_util.jit()
def compute_central_neighbors(pixel_neighbors, pixel_neighbors_size, shape, pixels):

    for x in range(1, shape[0] - 1):
        for y in range(1, shape[1] - 1):
            pixel_index = x * shape[1] + y
            pixel_neighbors[pixel_index, 0:4] = np.array([pixel_index - shape[1], pixel_index - 1,
                                                          pixel_index + 1, pixel_index + shape[1]])
            pixel_neighbors_size[pixel_index] = 4

    return pixel_neighbors, pixel_neighbors_size

@decorator_util.jit()
def voronoi_neighbors_from_pixels_and_ridge_points(pixels, ridge_points):
    """Compute the neighbors of every pixel as a list of the pixel index's each pixel shares a vertex with.

    The ridge points of the Voronoi grid are used to derive this.

    Parameters
    ----------
    ridge_points : scipy.spatial.Voronoi.ridge_points
        Each Voronoi-ridge (two indexes representing a pixel mapping_matrix).
    """

    pixel_neighbors_size = np.zeros(shape=(pixels))

    for ridge_index in range(ridge_points.shape[0]):
        pair0 = ridge_points[ridge_index, 0]
        pair1 = ridge_points[ridge_index, 1]
        pixel_neighbors_size[pair0] += 1
        pixel_neighbors_size[pair1] += 1

    pixel_neighbors_index = np.zeros(shape=(pixels))
    pixel_neighbors = -1 * np.ones(shape=(pixels, int(np.max(pixel_neighbors_size))))

    for ridge_index in range(ridge_points.shape[0]):
        pair0 = ridge_points[ridge_index, 0]
        pair1 = ridge_points[ridge_index, 1]
        pixel_neighbors[pair0, int(pixel_neighbors_index[pair0])] = pair1
        pixel_neighbors[pair1, int(pixel_neighbors_index[pair1])] = pair0
        pixel_neighbors_index[pair0] += 1
        pixel_neighbors_index[pair1] += 1

    return pixel_neighbors, pixel_neighbors_size


import numpy as np

from autofit.tools import fit
from autolens import exc
from autolens.lens.util import lens_fit_util as util
from autolens.model.inversion import inversions


class LensDataFit(fit.DataFit):
    def __init__(self, image, noise_map, mask, model_image):
        super().__init__(data=image, noise_map=noise_map, mask=mask, model_data=model_image)

    @property
    def model_image(self):
        return self.model_data

    @property
    def image(self):
        return self.data

    @property
    def figure_of_merit(self):
        return self.likelihood

    @classmethod
    def for_data_and_tracer(cls, lens_data, tracer, padded_tracer=None):
        """Fit lens data with a model tracer, automatically determining the type of fit based on the \
        properties of the galaxies in the tracer.

        Parameters
        -----------
        lens_data : lens_data.LensData or lens_data.LensDataHyper
            The lens-images that is fitted.
        tracer : ray_tracing.TracerNonStack
            The tracer, which describes the ray-tracing and strong lens configuration.
        padded_tracer : ray_tracing.Tracer or None
            A tracer with an identical strong lens configuration to the tracer above, but using the lens data's \
            padded grid_stack such that unmasked model-images can be computed.
        """

        if tracer.has_light_profile and not tracer.has_pixelization:
            return LensProfileFit(lens_data=lens_data, tracer=tracer, padded_tracer=padded_tracer)
        elif not tracer.has_light_profile and tracer.has_pixelization:
            return LensInversionFit(lens_data=lens_data, tracer=tracer, padded_tracer=None)
        elif tracer.has_light_profile and tracer.has_pixelization:
            return LensProfileInversionFit(lens_data=lens_data, tracer=tracer, padded_tracer=None)
        else:
            raise exc.FittingException('The fit routine did not call a Fit class - check the '
                                       'properties of the tracer')


class LensTracerFit(LensDataFit):

    def __init__(self, image, noise_map, mask, model_image, tracer, psf, map_to_scaled_array, padded_tracer=None):
        """ An  lens fitter, which contains the tracer's used to perform the fit and functions to manipulate \
        the lens data's hyper.

        Parameters
        -----------
        tracer : ray_tracing.Tracer
            The tracer, which describes the ray-tracing and strong lens configuration.
        padded_tracer : ray_tracing.TracerNonStack or None
            A tracer with an identical strong lens configuration to the tracer above, but using the lens data's \
            padded grid_stack such that unmasked model-images can be computed.
        map_to_scaled_array : func
            A function which maps the 1D lens hyper to its unmasked 2D array.
        """
        # Add hyper noise if these is a padded tracer...
        if padded_tracer is not None:
            original_noise_map = noise_map.copy()
            unmasked_model_image = padded_tracer.image_plane.grid_stack. \
                unmasked_blurred_image_from_psf_and_unmasked_image(psf=psf,
                                                                   unmasked_image_1d=padded_tracer.
                                                                   image_plane_image_1d)
            # ...to any galaxy with a corresponding HyperGalaxy
            for galaxy in tracer.galaxies:
                if galaxy.hyper_galaxy is not None:
                    plane = padded_tracer.plane_with_galaxy(galaxy)
                    unmasked_galaxy_image = plane.unmasked_blurred_image_of_galaxy_with_grid_stack_psf(
                        galaxy,
                        padded_tracer.image_plane.grid_stack,
                        psf)
                    hyper_galaxy = galaxy.hyper_galaxy
                    hyper_noise = hyper_galaxy.hyper_noise_from_model_image_galaxy_image_and_noise_map(
                        unmasked_model_image,
                        unmasked_galaxy_image,
                        original_noise_map)
                    noise_map += hyper_noise
        super().__init__(image=image, noise_map=noise_map, mask=mask, model_image=model_image)
        self.tracer = tracer
        self.padded_tracer = padded_tracer
        self.psf = psf
        self.map_to_scaled_array = map_to_scaled_array

    @property
    def total_inversions(self):
        return len(self.tracer.mappers_of_planes)

    @property
    def unmasked_model_image(self):
        if self.padded_tracer is not None and not self.padded_tracer.has_pixelization:
            return self.padded_tracer.image_plane.grid_stack.unmasked_blurred_image_from_psf_and_unmasked_image(
                psf=self.psf,
                unmasked_image_1d=self.padded_tracer.image_plane_image_1d)

    @property
    def unmasked_model_image_of_planes(self):
        if self.padded_tracer is not None:
            return util.unmasked_blurred_image_of_planes_from_padded_grid_stack_and_psf(
                planes=self.padded_tracer.planes, padded_grid_stack=self.padded_tracer.image_plane.grid_stack,
                psf=self.psf)

    def unmasked_model_image_for_galaxy(self, galaxy):
        plane = self.padded_tracer.plane_with_galaxy(galaxy)
        return plane.unmasked_blurred_image_of_galaxy_with_grid_stack_psf(galaxy,
                                                                          self.padded_tracer.image_plane.grid_stack,
                                                                          self.psf)

    @property
    def unmasked_model_image_of_planes_and_galaxies(self):
        if self.padded_tracer is not None:
            return util.unmasked_blurred_image_of_planes_and_galaxies_from_padded_grid_stack_and_psf(
                planes=self.padded_tracer.planes, padded_grid_stack=self.padded_tracer.image_plane.grid_stack,
                psf=self.psf)


class LensProfileFit(LensTracerFit):

    def __init__(self, lens_data, tracer, padded_tracer=None):
        """ An  lens profile fitter, which generates the image-plane image of all galaxies (with light \
        profiles) in the tracer and blurs it with the lens data's PSF.

        If a padded tracer is supplied, the blurred profile image's can be generated over the entire image and thus \
        without the mask.

        Parameters
        -----------
        lens_data : lens_data.LensData
            The lens-image that is fitted.
        tracer : ray_tracing.TracerNonStack
            The tracer, which describes the ray-tracing and strong lens configuration.
        padded_tracer : ray_tracing.Tracer or None
            A tracer with an identical strong lens configuration to the tracer above, but using the lens data's \
            padded grid_stack such that unmasked model-images can be computed.
        """
        blurred_profile_image_1d = util.blurred_image_1d_from_1d_unblurred_and_blurring_images(
            unblurred_image_1d=tracer.image_plane_image_1d, blurring_image_1d=tracer.image_plane_blurring_image_1d,
            convolver=lens_data.convolver_image)
        super(LensProfileFit, self).__init__(
            image=lens_data.image, noise_map=lens_data.noise_map,
            mask=lens_data.mask, model_image=lens_data.map_to_scaled_array(
                array_1d=blurred_profile_image_1d),
            tracer=tracer, padded_tracer=padded_tracer, psf=lens_data.psf,
            map_to_scaled_array=lens_data.map_to_scaled_array)

        self.convolver_image = lens_data.convolver_image

    @property
    def model_image_of_planes(self):
        return util.blurred_image_of_planes_from_1d_images_and_convolver(
            total_planes=self.tracer.total_planes,
            image_plane_image_1d_of_planes=self.tracer.image_plane_image_1d_of_planes,
            image_plane_blurring_image_1d_of_planes=self.tracer.image_plane_blurring_image_1d_of_planes,
            convolver=self.convolver_image,
            map_to_scaled_array=self.map_to_scaled_array)

    @property
    def figure_of_merit(self):
        return self.likelihood


class InversionFit(LensTracerFit):
    def __init__(self, lens_data, model_image, tracer, inversion, padded_tracer=None):
        super().__init__(image=lens_data.image, noise_map=lens_data.noise_map,
                         mask=lens_data.mask,
                         model_image=model_image,
                         psf=lens_data.psf,
                         tracer=tracer,
                         padded_tracer=padded_tracer,
                         map_to_scaled_array=lens_data.map_to_scaled_array)
        self.inversion = inversion

        self.likelihood_with_regularization = \
            util.likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
                chi_squared=self.chi_squared, regularization_term=inversion.regularization_term,
                noise_normalization=self.noise_normalization)

        self.evidence = util.evidence_from_inversion_terms(
            chi_squared=self.chi_squared,
            regularization_term=inversion.regularization_term,
            log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
            log_regularization_term=inversion.log_det_regularization_matrix_term,
            noise_normalization=self.noise_normalization)


class LensInversionFit(InversionFit):

    def __init__(self, lens_data, tracer, padded_tracer=None):
        """ An  lens inversion fitter, which fits the lens data an inversion using the mapper(s) and \
        regularization(s) in the galaxies of the tracer.

        This inversion use's the lens-image, its PSF and an input noise-map.

        Parameters
        -----------
        lens_data : lens_data.LensData
            The lens-image that is fitted.
        tracer : ray_tracing.Tracer
            The tracer, which describes the ray-tracing and strong lens configuration.
        """

        inversion = inversions.inversion_from_image_mapper_and_regularization(
            image_1d=lens_data.image_1d, noise_map_1d=lens_data.noise_map_1d,
            convolver=lens_data.convolver_mapping_matrix,
            mapper=tracer.mappers_of_planes[-1], regularization=tracer.regularizations_of_planes[-1])
        super().__init__(lens_data=lens_data,
                         model_image=inversion.reconstructed_data,
                         tracer=tracer,
                         inversion=inversion,
                         padded_tracer=padded_tracer)

    @property
    def figure_of_merit(self):
        return self.evidence

    @property
    def model_image_of_planes(self):
        return [None, self.inversion.reconstructed_data]


class LensProfileInversionFit(InversionFit):

    def __init__(self, lens_data, tracer, padded_tracer=None):
        """ An  lens profile and inversion fitter, which first generates and subtracts the image-plane \
        image of all galaxies (with light profiles) in the tracer, blurs it with the PSF and fits the residual image \
        with an inversion using the mapper(s) and regularization(s) in the galaxy's of the tracer.

        This inversion use's the lens-image, its PSF and an input noise-map.

        If a padded tracer is supplied, the blurred profile image's can be generated over the entire image and thus \
        without the mask.

        Parameters
        -----------
        lens_data : lens_data.LensData
            The lens-image that is fitted.
        tracer : ray_tracing.Tracer
            The tracer, which describes the ray-tracing and strong lens configuration.
        padded_tracer : ray_tracing.TracerNonStack or None
            A tracer with an identical strong lens configuration to the tracer above, but using the lens data's \
            padded grid_stack such that unmasked model-images can be computed.
        """
        blurred_profile_image_1d = util.blurred_image_1d_from_1d_unblurred_and_blurring_images(
            unblurred_image_1d=tracer.image_plane_image_1d, blurring_image_1d=tracer.image_plane_blurring_image_1d,
            convolver=lens_data.convolver_image)

        blurred_profile_image = lens_data.map_to_scaled_array(array_1d=blurred_profile_image_1d)
        profile_subtracted_image_1d = lens_data.image_1d - blurred_profile_image_1d
        inversion = inversions.inversion_from_image_mapper_and_regularization(
            image_1d=profile_subtracted_image_1d, noise_map_1d=lens_data.noise_map_1d,
            convolver=lens_data.convolver_mapping_matrix, mapper=tracer.mappers_of_planes[-1],
            regularization=tracer.regularizations_of_planes[-1])

        model_image = blurred_profile_image + inversion.reconstructed_data
        super(LensProfileInversionFit, self).__init__(tracer=tracer,
                                                      padded_tracer=padded_tracer,
                                                      lens_data=lens_data,
                                                      model_image=model_image,
                                                      inversion=inversion)

        self.convolver_image = lens_data.convolver_image

        self.blurred_profile_image = blurred_profile_image
        self.profile_subtracted_image = lens_data.image - self.blurred_profile_image

    @property
    def model_image_of_planes(self):
        return [self.blurred_profile_image, self.inversion.reconstructed_data]

    @property
    def figure_of_merit(self):
        return self.evidence


class LensPositionFit(object):

    def __init__(self, positions, noise_map):
        """A lens position fitter, which takes a set of positions (e.g. from a plane in the tracer) and computes \
        their maximum separation, such that points which tracer closer to one another have a higher likelihood.

        Parameters
        -----------
        positions : [[]]
            The (y,x) arc-second coordinates of positions which the maximum distance and likelihood is computed using.
        noise_map : ndarray | float
            The noise-value assumed when computing the likelihood.
        """
        self.positions = positions
        self.noise_map = noise_map

    @property
    def chi_squared_map(self):
        return np.square(np.divide(self.maximum_separations, self.noise_map))

    @property
    def figure_of_merit(self):
        return -0.5 * sum(self.chi_squared_map)

    def maximum_separation_within_threshold(self, threshold):
        return max(self.maximum_separations) <= threshold

    @property
    def maximum_separations(self):
        return list(map(lambda positions: self.max_separation_of_grid(positions), self.positions))

    @staticmethod
    def max_separation_of_grid(grid):
        rdist_max = np.zeros((grid.shape[0]))
        for i in range(grid.shape[0]):
            xdists = np.square(np.subtract(grid[i, 0], grid[:, 0]))
            ydists = np.square(np.subtract(grid[i, 1], grid[:, 1]))
            rdist_max[i] = np.max(np.add(xdists, ydists))
        return np.max(np.sqrt(rdist_max))

# TODO : The [plane_index][galaxy_index] datas structure is going to be key to tracking galaxies / hyper galaxies in
# TODO : Multi-plane ray tracing. I never felt it was easy to follow using list comprehensions from ray_tracing.
# TODO : Can we make this neater?



import os
import numpy as np

from autofit import conf
from autofit.optimize import non_linear as nl
from autolens.data.array import mask as msk
from autolens.model.galaxy import galaxy as g, galaxy_model as gm
from autolens.model.galaxy import galaxy_data as gd
from autolens.model.galaxy.util import galaxy_util
from autolens.data.array import grids, scaled_array
from autolens.pipeline import phase as ph
from autolens.model.profiles import mass_profiles as mp

# In this script, we're going to simulate a surface-density profile using a set of mass-profiles, and perform a direct
# fit to this surface density using a model galaxy. This uses a non-linear search and the fitting process is equivalent
# to fitting an image, but it bypasses the lens modeling process (e.g. the source reconstruction and usse of 'real'
# imaging data).

# Whilst this may sound like an odd thing to do, there are reasons why one may wish to perform a direct fit to a
# derived light or mass profile quantity, like the surface density:

# 1) If the mass-profile(s) used to generate the galaxy that is fitted and the model galaxy are different, this fit
#    will inform us of how the mismatch between profiles leads to a different estimate of the inferred mass profile
#    properties.

# 2) By bypassing the lens modeling process, we can test what results we get whilst bypass the potential systematics
#    that arise from a lens model fit (e.g due to the source reconstruction or quality of data).

# Setup the path to the workspace, using a relative directory name.
workspace_path = '{}/../../'.format(os.path.dirname(os.path.realpath(__file__)))

# Use this path to explicitly set the config path and output path.
conf.instance = conf.Config(config_path=workspace_path + 'config', output_path=workspace_path + 'output')

# First, we'll setup the grid stack we use to simulate a surface density profile.
pixel_scale = 0.05
image_shape = (250, 250)
grid_stack = grids.GridStack.from_shape_pixel_scale_and_sub_grid_size(shape=image_shape, pixel_scale=pixel_scale,
                                                                      sub_grid_size=4)

# Now lets create a galaxy, using a simple singular isothermal sphere.
galaxy = g.Galaxy(mass=mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0))

# Next, we'll generate its surface density profile. Note that, because we're using the galaxy_util surface density
# function, the sub-grid of the grid-stack that is passed to this function is used to over-sample the surface density.
# The surface density averages over this sub-grid, such that it is the shape of the image (250, 250).
surface_density = galaxy_util.convergence_of_galaxies_from_grid(galaxies=[galaxy], grid=grid_stack.sub)
surface_density = grid_stack.regular.scaled_array_from_array_1d(array_1d=surface_density)

# Now, we'll set this surface density up as our 'galaxy-data', meaning that it is what we'll fit via a non-linear
# search phase. To perform a fit we need a noise-map to help define our chi-squared. Given we are fitting a direct
# lensing quantity the actual values of this noise-map arn't particularly important, so we'll just use a noise-map of
# all 0.1's
noise_map = scaled_array.ScaledSquarePixelArray(array=0.1*np.ones(surface_density.shape), pixel_scale=pixel_scale)
data = gd.GalaxyData(image=surface_density, noise_map=noise_map, pixel_scale=pixel_scale)

# The fit will use a mask, which we setup like any other fit. Lets use a circular mask of 2.0"
def mask_function_circular(image):
    return msk.Mask.circular(shape=image.shape, pixel_scale=image.pixel_scale, radius_arcsec=2.0)

# We can now use a special phase, called a 'GalaxyFitPhase', to fit this surface density with a model galaxy the
# mass-profiles of which we get to choose. We'll fit it with a singular isothermal sphere and should see we retrieve
# the input model above.
phase = ph.GalaxyFitPhase(phase_name='galaxy_surface_density_fit',
                          galaxies=dict(lens=gm.GalaxyModel(mass=mp.SphericalIsothermal)), use_surface_density=True,
                          sub_grid_size=4, mask_function=mask_function_circular,
                          optimizer_class=nl.MultiNest)

phase.run(galaxy_data=[data])

# If you check your output folder, you should see that this fit has been performed and visualization specific to a
# surface-density fit is output.

# Fits to an intensity map and gravitational potential are also possible. To do so, simply change the profile quantitiy
# that is simuulated and edit the 'use_surface_density' flag in the GalaxyFitPhase above to the appropriate quantity.
# You can also fit deflection angle maps - however this requires a few small changes to this script, so we've create a
# 'galaxy_fit_deflections.py' example script to show you how.


from autolens.data.array import scaled_array
from autolens.data.array.util import array_util
from autolens.data.array.plotters import array_plotters

import os

# In this example, we will load the image of a strong lens from a .fits file and plot it using the
# function autolens.data.plotters.array_plotters.plot_array. We will customize the appearance of this figure to
# highlight the features of the image. For more generical plotting tools (e.g. changing the figure size, axis units,
# outputting the image to the hard-disk, etc.) checkout the example in 'workspace/plotting/examples/arrays/array.py'.

# We will use the image of slacs1430+4105.

# Setup the path to the workspace, using a relative directory name.
workspace_path = '{}/../../../'.format(os.path.dirname(os.path.realpath(__file__)))

# We have included the .fits data required for this example in the directory
# 'workspace/output/data/example/slacs1430+4105/'.

# First, lets setup the path to the .fits file of the image.
data_type = 'example'
data_name = 'slacs1430+4105'

# Create the path where the data will be loaded from, which in this case is
# '/workspace/data/example/slacs1430+4105/'
data_path = array_util.make_and_return_path(path=workspace_path, folder_names=['data', data_type, data_name])
image_path = data_path + 'image.fits'

# Now, lets load this array as a scaled array. A scaled array is an ordinary NumPy array, but it also includes a pixel
# scale which allows us to convert the axes of the array to arc-second coordinates.
image = scaled_array.ScaledSquarePixelArray.from_fits_with_pixel_scale(file_path=image_path, hdu=0, pixel_scale=0.03)

# We can now use an array plotter to plot the image. Lets first plot it using the default PyAutoLens settings.
array_plotters.plot_array(array=image, title='SLACS1430+4105 Image')

# For a lens like SLACS1430+4105, the lens galaxy's light outshines the background source, making it appear faint.
# we can use a symmetric logarithmic colorbar normalization to better reveal the source galaxy (due to negative values
# in the image, we cannot use a regular logirithmic colorbar normalization).
array_plotters.plot_array(array=image, title='SLACS1430+4105 Image',
                          norm='symmetric_log', linthresh=0.05, linscale=0.02)

# Alternatively, we can use the default linear colorbar normalization and customize the limits over which the colormap
# spans its dynamic range.
array_plotters.plot_array(array=image, title='SLACS1430+4105 Image',
                          norm='linear', norm_min=0.0, norm_max=0.3)

# We can also load the full set of ccd data (image, noise-map, PSF) and use the ccd_plotters to make the figures above.
from autolens.data import ccd
from autolens.data.plotters import ccd_plotters

psf_path = workspace_path + '/data/example/' + data_name + '/psf.fits'
noise_map_path = workspace_path + '/data/example/' + data_name + '/noise_map.fits'

ccd_data = ccd.load_ccd_data_from_fits(image_path=data_path, psf_path=psf_path, noise_map_path=noise_map_path,
                                       pixel_scale=0.03)

# These plotters can be customized using the exact same functions as above.
ccd_plotters.plot_noise_map(ccd_data=ccd_data, title='SLACS1430+4105 Noise-Map',
                            norm='log')

# Of course, as we've seen in many other examples, a sub-plot of the ccd data can be plotted. This can also take the
# customization inputs above, but it should be noted that the options are applied to all images, and thus will most
# likely degrade a number of the sub-plot images.
ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data, norm='symmetric_log', linthresh=0.05, linscale=0.02)


import numpy as np
import pytest

from autolens.data import ccd, convolution
from autolens.data.array.util import grid_util
from autolens.data.array import grids
from autolens.data.array import scaled_array
from autolens.data.array import mask as msk
from autolens.lens import lens_data as ld
from autolens.model.inversion import convolution as inversion_convolution


@pytest.fixture(name='ccd')
def make_ccd():

    image = scaled_array.ScaledSquarePixelArray(array=np.ones((6, 6)), pixel_scale=3.0)
    psf = ccd.PSF(array=np.ones((3, 3)), pixel_scale=3.0, renormalize=False)
    noise_map = ccd.NoiseMap(array=2.0 * np.ones((6, 6)), pixel_scale=3.0)
    background_noise_map = ccd.NoiseMap(array=3.0 * np.ones((6, 6)), pixel_scale=3.0)
    poisson_noise_map = ccd.PoissonNoiseMap(array=4.0 * np.ones((6, 6)), pixel_scale=3.0)
    exposure_time_map = ccd.ExposureTimeMap(array=5.0 * np.ones((6, 6)), pixel_scale=3.0)
    background_sky_map = scaled_array.ScaledSquarePixelArray(array=6.0 * np.ones((6, 6)), pixel_scale=3.0)

    return ccd.CCDData(image=image, pixel_scale=3.0, psf=psf, noise_map=noise_map,
                       background_noise_map=background_noise_map, poisson_noise_map=poisson_noise_map,
                       exposure_time_map=exposure_time_map, background_sky_map=background_sky_map)

@pytest.fixture(name="mask")
def make_mask():
    return msk.Mask(np.array([[True, True, True, True, True, True],
                              [True, True, True, True, True, True],
                              [True, True, False, False, True, True],
                              [True, True, False, False, True, True],
                              [True, True, True, True, True, True],
                              [True, True, True, True, True, True]]), pixel_scale=3.0)

@pytest.fixture(name="lens_data")
def make_lens_ccd(ccd, mask):
    return ld.LensData(ccd_data=ccd, mask=mask)


class TestLensData(object):

    def test__attributes(self, ccd, lens_data):

        assert lens_data.pixel_scale == ccd.pixel_scale
        assert lens_data.pixel_scale == 3.0

        assert (lens_data.image == ccd.image).all()
        assert (lens_data.image == np.ones((6,6))).all()

        assert (lens_data.psf == ccd.psf).all()
        assert (lens_data.psf == np.ones((3,3))).all()

        assert (lens_data.noise_map == ccd.noise_map).all()
        assert (lens_data.noise_map == 2.0*np.ones((6,6))).all()

        assert lens_data.image_psf_shape == (3,3)
        assert lens_data.inversion_psf_shape == (3,3)

    def test__masking(self, lens_data):

        assert (lens_data.image_1d == np.ones(4)).all()
        assert (lens_data.noise_map_1d == 2.0*np.ones(4)).all()
        assert (lens_data.mask_1d == np.array([False, False, False, False])).all()

    def test__grid_stack(self, lens_data):

        assert (lens_data.grid_stack.regular == np.array([[1.5, -1.5], [1.5, 1.5], [-1.5, -1.5], [-1.5, 1.5]])).all()
        assert (lens_data.grid_stack.sub == np.array([[2.0, -2.0], [2.0, -1.0], [1.0, -2.0], [1.0, -1.0],
                                                     [2.0, 1.0], [2.0, 2.0], [1.0, 1.0], [1.0, 2.0],
                                                     [-1.0, -2.0], [-1.0, -1.0], [-2.0, -2.0], [-2.0, -1.0],
                                                     [-1.0, 1.0], [-1.0, 2.0], [-2.0, 1.0], [-2.0, 2.0]])).all()
        assert (lens_data.grid_stack.blurring == np.array([[4.5, -4.5], [4.5, -1.5], [4.5, 1.5], [4.5, 4.5],
                                                          [1.5, -4.5], [1.5, 4.5], [-1.5, -4.5], [-1.5, 4.5],
                                                          [-4.5, -4.5], [-4.5, -1.5], [-4.5, 1.5], [-4.5, 4.5]])).all()

    def test__padded_grid_stack(self, lens_data):

        padded_image_util = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(mask=np.full((8, 8), False),
                                                                        pixel_scales=lens_data.image.pixel_scales)

        assert (lens_data.padded_grid_stack.regular == padded_image_util).all()
        assert lens_data.padded_grid_stack.regular.image_shape == (6, 6)
        assert lens_data.padded_grid_stack.regular.padded_shape == (8, 8)

        padded_sub_util = grid_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(
            mask=np.full((8, 8), False), pixel_scales=lens_data.image.pixel_scales,
            sub_grid_size=lens_data.grid_stack.sub.sub_grid_size)

        assert lens_data.padded_grid_stack.sub == pytest.approx(padded_sub_util, 1e-4)
        assert lens_data.padded_grid_stack.sub.image_shape == (6, 6)
        assert lens_data.padded_grid_stack.sub.padded_shape == (8, 8)

        assert (lens_data.padded_grid_stack.blurring == np.array([[0.0, 0.0]])).all()

    def test__interp_pixel_scale_input__grid_stack_and_padded_grid_stack_include_interpolators(self, ccd, mask):

        lens_data = ld.LensData(ccd_data=ccd, mask=mask, interp_pixel_scale=1.0)

        grid_stack = grids.GridStack.grid_stack_from_mask_sub_grid_size_and_psf_shape(
                        mask=mask, sub_grid_size=2, psf_shape=(3, 3))
        new_grid_stack = grid_stack.new_grid_stack_with_interpolator_added_to_each_grid(interp_pixel_scale=1.0)

        assert (lens_data.grid_stack.regular == new_grid_stack.regular).all()
        assert (lens_data.grid_stack.regular.interpolator.vtx == new_grid_stack.regular.interpolator.vtx).all()
        assert (lens_data.grid_stack.regular.interpolator.wts == new_grid_stack.regular.interpolator.wts).all()

        assert (lens_data.grid_stack.sub == new_grid_stack.sub).all()
        assert (lens_data.grid_stack.sub.interpolator.vtx == new_grid_stack.sub.interpolator.vtx).all()
        assert (lens_data.grid_stack.sub.interpolator.wts == new_grid_stack.sub.interpolator.wts).all()
        
        assert (lens_data.grid_stack.blurring == new_grid_stack.blurring).all()
        assert (lens_data.grid_stack.blurring.interpolator.vtx == new_grid_stack.blurring.interpolator.vtx).all()
        assert (lens_data.grid_stack.blurring.interpolator.wts == new_grid_stack.blurring.interpolator.wts).all()

        padded_grid_stack = grids.GridStack.padded_grid_stack_from_mask_sub_grid_size_and_psf_shape(
                        mask=mask, sub_grid_size=2, psf_shape=(3, 3))
        new_padded_grid_stack = \
            padded_grid_stack.new_grid_stack_with_interpolator_added_to_each_grid(interp_pixel_scale=1.0)

        assert (lens_data.padded_grid_stack.regular == new_padded_grid_stack.regular).all()
        assert (lens_data.padded_grid_stack.regular.interpolator.vtx == new_padded_grid_stack.regular.interpolator.vtx).all()
        assert (lens_data.padded_grid_stack.regular.interpolator.wts == new_padded_grid_stack.regular.interpolator.wts).all()

        assert (lens_data.padded_grid_stack.sub == new_padded_grid_stack.sub).all()
        assert (lens_data.padded_grid_stack.sub.interpolator.vtx == new_padded_grid_stack.sub.interpolator.vtx).all()
        assert (lens_data.padded_grid_stack.sub.interpolator.wts == new_padded_grid_stack.sub.interpolator.wts).all()

        assert (lens_data.padded_grid_stack.blurring == np.array([[0.0, 0.0]])).all()

    def test__border(self, lens_data):
        assert (lens_data.border == np.array([0, 1, 2, 3])).all()

    def test__convolvers(self, lens_data):
        assert type(lens_data.convolver_image) == convolution.ConvolverImage
        assert type(lens_data.convolver_mapping_matrix) == inversion_convolution.ConvolverMappingMatrix

    def test__constructor_inputs(self):

        psf = ccd.PSF(np.ones((7, 7)), 1)
        image = ccd.CCDData(np.ones((51, 51)), pixel_scale=3., psf=psf, noise_map=np.ones((51, 51)))
        mask = msk.Mask.unmasked_for_shape_and_pixel_scale(shape=(51, 51), pixel_scale=1.0, invert=True)
        mask[26, 26] = False

        lens_data = ld.LensData(image, mask, sub_grid_size=8, image_psf_shape=(5, 5),
                                inversion_psf_shape=(3, 3), positions=[np.array([[1.0, 1.0]])])

        assert lens_data.sub_grid_size == 8
        assert lens_data.convolver_image.psf_shape == (5, 5)
        assert lens_data.convolver_mapping_matrix.psf_shape == (3, 3)
        assert (lens_data.positions[0] == np.array([[1.0, 1.0]])).all()

        assert lens_data.image_psf_shape == (5,5)
        assert lens_data.inversion_psf_shape == (3, 3)

    def test__lens_data_with_modified_image(self, lens_data):

        lens_data = lens_data.new_lens_data_with_modified_image(modified_image=8.0 * np.ones((6, 6)))

        assert (lens_data.image == 8.0*np.ones((6,6))).all()
        assert (lens_data.image_1d == 8.0*np.ones(4)).all()

    def test__lens_data_with_binned_up_ccd_data(self, lens_data):

        binned_up_psf = lens_data.ccd_data.psf.new_psf_with_rescaled_odd_dimensioned_array(rescale_factor=0.5)

        lens_data = lens_data.new_lens_data_with_binned_up_ccd_data_and_mask(bin_up_factor=2)

        assert (lens_data.image == np.ones((3,3))).all()
        assert (lens_data.psf == binned_up_psf).all()
        assert (lens_data.noise_map == np.ones((3,3))).all()
        assert (lens_data.ccd_data.background_noise_map == (np.sqrt(36)/4.0)*np.ones((3,3))).all()
        assert (lens_data.ccd_data.poisson_noise_map == (np.sqrt(64)/4.0)*np.ones((3,3))).all()
        assert (lens_data.ccd_data.exposure_time_map == 20.0*np.ones((3,3))).all()
        assert (lens_data.ccd_data.background_sky_map == 6.0*np.ones((3,3))).all()

        assert (lens_data.mask == np.array([[True, True, True],
                                            [True, False, True],
                                            [True, True, True]])).all()

        assert (lens_data.image_1d == np.ones((1))).all()
        assert (lens_data.noise_map_1d == np.ones((1))).all()

@pytest.fixture(name="lens_data_hyper")
def make_lens_hyper_image(ccd, mask):

    return ld.LensDataHyper(ccd_data=ccd, mask=mask, hyper_model_image=10.0 * np.ones((6, 6)),
                            hyper_galaxy_images=[11.0*np.ones((6,6)), 12.0*np.ones((6,6))],
                            hyper_minimum_values=[0.1, 0.2])


class TestLensDataHyper(object):

    def test__attributes(self, ccd, lens_data_hyper):

        assert lens_data_hyper.pixel_scale == ccd.pixel_scale

        assert (lens_data_hyper.image == ccd.image).all()
        assert (lens_data_hyper.image == np.ones((6,6))).all()

        assert (lens_data_hyper.psf == ccd.psf).all()
        assert (lens_data_hyper.psf == np.ones((3,3))).all()

        assert (lens_data_hyper.noise_map == ccd.noise_map).all()
        assert (lens_data_hyper.noise_map == 2.0*np.ones((6,6))).all()

        assert (lens_data_hyper.hyper_model_image == 10.0*np.ones((6,6))).all()
        assert (lens_data_hyper.hyper_galaxy_images[0] == 11.0*np.ones((6,6))).all()
        assert (lens_data_hyper.hyper_galaxy_images[1] == 12.0*np.ones((6,6))).all()

        assert lens_data_hyper.hyper_minimum_values == [0.1, 0.2]

    def test__masking(self, lens_data_hyper):

        assert (lens_data_hyper.image_1d == np.ones(4)).all()
        assert (lens_data_hyper.noise_map_1d == 2.0*np.ones(4)).all()

        assert (lens_data_hyper.hyper_model_image_1d == 10.0*np.ones(4)).all()
        assert (lens_data_hyper.hyper_galaxy_images_1d[0] == 11.0*np.ones(4)).all()
        assert (lens_data_hyper.hyper_galaxy_images_1d[1] == 12.0*np.ones(4)).all()



import sys
import math
import numpy as np
import matplotlib.pyplot as plt

sys.path.append("../")

from paper_plots import density_plot_tools
from autolens.analysis import galaxy
from autolens.model.profiles import mass_profiles


ltm_skip = 0
center_skip = 0

image_dir = '/gpfs/data_vector/pdtw24/PL_Data/SL03_2/'  # Dir of Object to make evidence tables from

image_name = 'SLACSJ0252+0039'
image_name = 'SLACSJ1250+0523'
# image_name = 'SLACSJ1430+4105'

slacs = density_plot_tools.SLACS(image_dir, image_name)

# pipeline_folder = 'Pipeline_Total'
# phase_folder = 'PL_Phase_8'

# pipeline_folder = 'Pipeline_LMDM'
# phase_folder = 'PL_Phase_7'

pipeline_folder = 'Pipeline_LTM2'
phase_folder = 'PL_5_LMDM_LTM_R_MC'

sersic_bulge = mass_profiles.EllipticalSersicMass(centre=(0.02, 0.008), axis_ratio=0.87,
                                                  phi=71.1, intensity=0.09458,
                                                  effective_radius=1.0153,
                                                  sersic_index=4.57468,
                                                  mass_to_light_ratio=1.55265)

exponential_halo = mass_profiles.EllipticalSersicMassRadialGradient(centre=(0.043, 0.036),
                                                                    axis_ratio=0.7861, phi=161.34, intensity=0.02475, effective_radius=2.1389, mass_to_light_ratio=1.2556,
                                                                    mass_to_light_gradient=-0.2229)

dark_matter_halo = mass_profiles.SphericalNFW(kappa_s=0.1035, scale_radius=30.0 * slacs.arcsec_per_kpc)

lens_galaxy = galaxy.Galaxy(redshift=slacs.redshift, mass_profiles=[sersic_bulge, exponential_halo, dark_matter_halo])
source_galaxy = galaxy.Galaxy(redshift=slacs.source_redshift)

number_bins = 30
radius_kpc = 30

radii = list(np.linspace(5e-3, radius_kpc * slacs.arcsec_per_kpc, number_bins + 1))

radii_plot = []
bulge_densities = []
halo_densities = []
dark_densities = []

for r in range(number_bins):

    annuli_area = (math.pi * radii[r + 1] ** 2 - math.pi * radii[r] ** 2)

    densities = ((lens_galaxy.dimensionless_mass_within_circles_individual(radii[r + 1]) -
                  lens_galaxy.dimensionless_mass_within_circles_individual(radii[r])) / annuli_area)

    bulge_densities.append(densities[0])
    halo_densities.append(densities[1])
    dark_densities.append(densities[2])

    radii_plot.append(((radii[r + 1] + radii[r]) / 2.0) * slacs.kpc_per_arcsec)

print(bulge_densities)

plt.semilogy(radii_plot, bulge_densities, color='r', label='Sersic Bulge')
plt.semilogy(radii_plot, halo_densities, color='g', label='EllipticalExponentialMass Halo')
plt.semilogy(radii_plot, dark_densities, color='k', label='Dark Matter Halo')

plt.legend()
plt.show()

#slacs.weighted_densities_vs_radii(radius_kpc=30.0, weight_cut=1e-3, number_bins=20)

#slacs.plot_density(image_name=image_name, labels=['Sersic', 'Exponential', 'NFWSph'])


import os

from autofit.tools import path_util
from autolens.data import ccd

test_path = '{}/../'.format(os.path.dirname(os.path.realpath(__file__)))


def pixel_scale_from_data_resolution(data_resolution):
    if data_resolution == 'LSST':
        return 0.2
    elif data_resolution == 'Euclid':
        return 0.1
    elif data_resolution == 'HST':
        return 0.05
    elif data_resolution == 'HST_Up':
        return 0.03
    elif data_resolution == 'AO':
        return 0.01
    else:
        raise ValueError('An invalid data resolution was entered - ', data_resolution)


def shape_from_data_resolution(data_resolution):
    if data_resolution == 'LSST':
        return 100, 100
    elif data_resolution == 'Euclid':
        return 150, 150
    elif data_resolution == 'HST':
        return 250, 250
    elif data_resolution == 'HST_Up':
        return 320, 320
    elif data_resolution == 'AO':
        return 750, 750
    else:
        raise ValueError('An invalid data-type was entered - ', data_resolution)


def data_resolution_from_pixel_scale(pixel_scale):
    if pixel_scale == 0.2:
        return 'LSST'
    elif pixel_scale == 0.1:
        return 'Euclid'
    elif pixel_scale == 0.05:
        return 'HST'
    elif pixel_scale == 0.03:
        return 'HST_Up'
    elif pixel_scale == 0.01:
        return 'AO'
    else:
        raise ValueError('An invalid pixel-scale was entered - ', pixel_scale)


def load_test_ccd_data(data_type, data_resolution, psf_shape=(11, 11), lens_name=None):
    pixel_scale = pixel_scale_from_data_resolution(data_resolution=data_resolution)

    data_path = path_util.make_and_return_path_from_path_and_folder_names(
        path=test_path, folder_names=['data', data_type, data_resolution])

    return ccd.load_ccd_data_from_fits(image_path=data_path + '/image.fits',
                                       psf_path=data_path + '/psf.fits',
                                       noise_map_path=data_path + '/noise_map.fits',
                                       pixel_scale=pixel_scale, resized_psf_shape=psf_shape, lens_name=lens_name)



from autolens.lens.summary import plane_summary

def summarize_tracer(summary_file, tracer, radii):

    # TODO : There will be methods here which summarize overall plane properties (e.g. total mass for all galaxies
    # TODO : redshifts. For now, we'll just iterate over all mass profiles and output the mass profile info.
    # TODO : Also not sure how to access the Galaxy name...

    summary_file.write('Tracer Cosmology = ' + str(tracer.cosmology) + '\n')
    summary_file.write('Tracer Redshifts = ' + str(tracer.plane_redshifts) + '\n\n')

    for plane in tracer.planes:

        plane_summary.summarize_plane(summary_file=summary_file, plane=plane,
                                      critical_surface_mass_density=tracer.critical_surface_mass_density_arcsec,
                                      radii=radii)


from autolens.model.galaxy.summary import galaxy_summary

def summarize_plane(summary_file, plane, critical_surface_mass_density, radii):

    # TODO : There will be methods here which summarize overall plane properties (e.g. total mass for all galaxies
    # TODO : redshifts. For now, we'll just iterate over all mass profiles and output the mass profile info.
    # TODO : Also not sure how to access the Galaxy name...

    summary_file.write('Plane Redshift = ' + str(plane.redshift) + '\n')
    summary_file.write('Plane Critical Surface Mass Density (solMass / arcsec^2) = ' + str(critical_surface_mass_density) + '\n\n')

    for galaxy in plane.galaxies:

        galaxy_summary.summarize_galaxy(summary_file=summary_file, galaxy=galaxy,
                                        critical_surface_mass_density=critical_surface_mass_density,
                                        cosmic_average_mass_density_arcsec=plane.cosmic_average_mass_density_arcsec,
                                        radii=radii)


from autolens.model.profiles import light_profiles
from autolens.model.profiles import mass_profiles
from autolens.model.galaxy import galaxy
from autolens.data.array import grids
from autolens.model.galaxy.plotters import galaxy_plotters

# In this example, we'll use 'mass_profiles' and 'light_profiles' modules introduced previously, along with the
# 'galaxy' module to create Galaxy objects in PyAutoLens. We'll see that:

# 1) Galaxies can take multiple light-profiles / mass-profiles and are composed of multiple components.
# 2) By taking multiple components, the combined intensities / deflection angle's of the profiles are computed.

# Lets use an identical grid to the previous example.
grid_stack = grids.GridStack.from_shape_pixel_scale_and_sub_grid_size(shape=(100, 100), pixel_scale=0.05,
                                                                      sub_grid_size=2)

# Lets make a galaxy with a Sersic light profile, by making a Sersic light profile and pasing it to a Galaxy object.
sersic_light_profile = light_profiles.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0,
                                                       intensity=1.0, effective_radius=1.0, sersic_index=2.5)
galaxy_with_light_profile = galaxy.Galaxy(light=sersic_light_profile)

# We can print the galaxy to confirm its profile and its parameters
print(galaxy_with_light_profile)

# In the previous example, we passed grids to the light-profile module to compute its image.
# We can do the exact same with galaxies, to again compute the galaxy's image.
galaxy_intensities = galaxy_with_light_profile.intensities_from_grid(grid=grid_stack.regular)
print('intensity of regular-grid pixel 1:')
print(galaxy_intensities[0])
print('intensity of regular-grid pixel 2:')
print(galaxy_intensities[1])
print('intensity of regular-grid pixel 3:')
print(galaxy_intensities[2])
print('etc.')

# A galaxy plotter allows us to the plot the image, just like the profile plotters did for a light
# profile (again, mapping the 1D image to 2D).
galaxy_plotters.plot_intensities(galaxy=galaxy_with_light_profile, grid=grid_stack.regular)

# We can pass galaxies as many profiles as we like. Lets create a galaxy with three light profiles.
light_profile_1 = light_profiles.SphericalSersic(centre=(0.0,  0.0), intensity=1.0, effective_radius=1.0, sersic_index=2.5)
light_profile_2 = light_profiles.SphericalSersic(centre=(1.0,  1.0), intensity=1.0, effective_radius=2.0, sersic_index=3.0)
light_profile_3 = light_profiles.SphericalSersic(centre=(1.0, -1.0), intensity=1.0, effective_radius=2.0, sersic_index=2.0)
galaxy_with_3_light_profiles = galaxy.Galaxy(light_1=light_profile_1, light_2=light_profile_2, light_3=light_profile_3)

# We can print the galaxy to confirm it possesses the Sersic light-profiles above.
print(galaxy_with_3_light_profiles)

# If we plot the galaxy, we see 3 blobs of light!
galaxy_plotters.plot_intensities(galaxy=galaxy_with_3_light_profiles, grid=grid_stack.regular)

# We can also plot each individual light profile using the 'subplot' galaxy plotter.
galaxy_plotters.plot_intensities_subplot(galaxy=galaxy_with_3_light_profiles, grid=grid_stack.regular)

# Mass profiles interact with Galaxy objects in the exact same way as light profiles.
# Lets create a galaxy with three SIS mass profiles.
mass_profile_1 = mass_profiles.SphericalIsothermal(centre=(0.0,  0.0), einstein_radius=1.0)
mass_profile_2 = mass_profiles.SphericalIsothermal(centre=(1.0,  1.0), einstein_radius=1.0)
mass_profile_3 = mass_profiles.SphericalIsothermal(centre=(1.0, -1.0), einstein_radius=1.0)
galaxy_with_3_mass_profiles = galaxy.Galaxy(mass_1=mass_profile_1, mass_2=mass_profile_2, mass_3=mass_profile_3)

# We can print a galaxy to confirm it possesses the sis mass-profiles above.
print(galaxy_with_3_mass_profiles)

# We can use a galaxy plotter to plot these deflection angles.

# (Deflection angles of mass-profiles add together just like the light-profile image's above)
galaxy_plotters.plot_deflections_y(galaxy=galaxy_with_3_mass_profiles, grid=grid_stack.regular)
galaxy_plotters.plot_deflections_x(galaxy=galaxy_with_3_mass_profiles, grid=grid_stack.regular)

# I wonder what 3 summed convergence profiles or potential's look like ;)
galaxy_plotters.plot_convergence(galaxy=galaxy_with_3_mass_profiles, grid=grid_stack.regular)
galaxy_plotters.plot_potential(galaxy=galaxy_with_3_mass_profiles, grid=grid_stack.regular)

# Finally, a galaxy can take both light and mass profiles, and there is no limit to how many we pass it.
light_profile_1 = light_profiles.SphericalSersic(centre=(0.0, 0.0), intensity=1.0,
                                                 effective_radius=1.0, sersic_index=1.0)
light_profile_2 = light_profiles.SphericalSersic(centre=(1.0, 1.0), intensity=1.0,
                                                 effective_radius=2.0, sersic_index=2.0)
light_profile_3 = light_profiles.SphericalSersic(centre=(2.0, 2.0), intensity=1.0,
                                                 effective_radius=3.0, sersic_index=3.0)
light_profile_4 = light_profiles.EllipticalSersic(centre=(1.0, -1.0), axis_ratio=0.5, phi=45.0,
                                                  intensity=1.0, effective_radius=1.0, sersic_index=1.0)

mass_profile_1 = mass_profiles.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0)
mass_profile_2 = mass_profiles.SphericalIsothermal(centre=(1.0, 1.0), einstein_radius=2.0)
mass_profile_3 = mass_profiles.SphericalIsothermal(centre=(2.0, 2.0), einstein_radius=3.0)
mass_profile_4 = mass_profiles.EllipticalIsothermal(centre=(1.0, -1.0), axis_ratio=0.5, phi=45.0,
                                                   einstein_radius=2.0)

galaxy_with_many_profiles = galaxy.Galaxy(light_1=light_profile_1, light_2=light_profile_2,
                                          light_3=light_profile_3, light_4=light_profile_4,
                                           mass_1=mass_profile_1, mass_2=mass_profile_2,
                                           mass_3=mass_profile_3, mass_4=mass_profile_4)

# Suffice to say, the galaxy's images, convergence, potential and deflections look pretty
# interesting.

galaxy_plotters.plot_intensities(galaxy=galaxy_with_many_profiles, grid=grid_stack.regular)
galaxy_plotters.plot_convergence(galaxy=galaxy_with_many_profiles, grid=grid_stack.regular)
galaxy_plotters.plot_potential(galaxy=galaxy_with_many_profiles, grid=grid_stack.regular)
galaxy_plotters.plot_deflections_y(galaxy=galaxy_with_many_profiles, grid=grid_stack.regular)
galaxy_plotters.plot_deflections_x(galaxy=galaxy_with_many_profiles, grid=grid_stack.regular)

# And we're done. Lets finished by just thinking about one question:

# 1) We've learnt we can group profiles into galaxies, to essentially sum the contribution of each light profile to the
#    galaxy image's intensity, or sum the contribution of each mass profile to the convergence, potential and
#    deflection angles. In strong lensing, there are often multiple galaxies next to one another responsible for the
#    lensing - how might we account for this?


from autolens.model.profiles.summary import profile_summary

from test.unit.mock.mock_summary import MockMassProfile, MockNFW, MockTruncatedNFW

import os

test_summary_dir = "{}/../../../test_files/summary/".format(os.path.dirname(os.path.realpath(__file__)))

class TestSummarizeMassProfile:

    def test__summarize_mass_profile(self):

        summary_file = open(file=test_summary_dir + 'model.summary', mode="w+")

        profile_summary.summarize_mass_profile(summary_file=summary_file, mass_profile=MockTruncatedNFW(),
                                               critical_surface_mass_density=1.0, cosmic_average_mass_density_arcsec=1.0,
                                               radii=[10.0, 500.0])

        summary_file.close()

        summary_file_read = open(test_summary_dir + 'model.summary', mode="r")
        summary_text = summary_file_read.readlines()

        assert summary_text == ['Mass Profile = MockTruncatedNFW\n',
                                '\n',
                                'Einstein Radius = 10.00"\n',
                                'Mass within Einstein Radius = 1.0000e+03 solMass\n',
                                'Mass within 10.00" = 1.0000e+03 solMass\n',
                                'Mass within 500.00" = 1.0000e+03 solMass\n',
                                'Rho at scale radius = 100.00\n',
                                'Delta concentration = 200.00\n',
                                'Concentration = 300.00\n',
                                'Radius at 200x cosmic average density = 400.00"\n',
                                'Mass at 200x cosmic average density = 500.00 solMass\n',
                                'Mass at truncation radius = 600.00 solMass\n']

        summary_file.close()

        os.remove(path=test_summary_dir + 'model.summary')

    def test__summarize_einstein_radius_and_mass___multiple_radii_both_output(self):

        summary_file = open(file=test_summary_dir + 'model.summary', mode="w+")

        profile_summary.summarize_einstein_radius_and_mass(summary_file=summary_file, mass_profile=MockMassProfile(),
                                                           critical_surface_mass_density=1.0)

        summary_file.close()

        summary_file_read = open(test_summary_dir + 'model.summary', mode="r")
        summary_text = summary_file_read.readlines()

        assert summary_text == ['Einstein Radius = 10.00"\n',
                                'Mass within Einstein Radius = 1.0000e+03 solMass\n']

        summary_file.close()

        os.remove(path=test_summary_dir + 'model.summary')

    def test__summarize_mass_within_radii___multiple_radii_both_output(self):

        summary_file = open(file=test_summary_dir + 'model.summary', mode="w+")

        profile_summary.summarize_mass_within_radii(summary_file=summary_file, mass_profile=MockMassProfile(),
                                                    radii=[10.0, 500.0], critical_surface_mass_density=1.0)

        summary_file.close()

        summary_file_read = open(test_summary_dir + 'model.summary', mode="r")
        summary_text = summary_file_read.readlines()

        assert summary_text == ['Mass within 10.00" = 1.0000e+03 solMass\n',
                                'Mass within 500.00" = 1.0000e+03 solMass\n']

        summary_file.close()

        os.remove(path=test_summary_dir + 'model.summary')

    def test__summarize_nfw_mass_profile(self):

        summary_file = open(file=test_summary_dir + 'model.summary', mode="w+")

        profile_summary.summarize_nfw_mass_profile(summary_file=summary_file, nfw=MockNFW(),
                                                   critical_surface_mass_density_arcsec=1.0,
                                                   cosmic_average_mass_density_arcsec=1.0)

        summary_file.close()

        summary_file_read = open(test_summary_dir + 'model.summary', mode="r")
        summary_text = summary_file_read.readlines()

        assert summary_text == ['Rho at scale radius = 100.00\n',
                                'Delta concentration = 200.00\n',
                                'Concentration = 300.00\n',
                                'Radius at 200x cosmic average density = 400.00"\n',
                                'Mass at 200x cosmic average density = 500.00 solMass\n']

        summary_file.close()

        os.remove(path=test_summary_dir + 'model.summary')

    def test__summarize_truncated_nfw_mass_profile(self):

        summary_file = open(file=test_summary_dir + 'model.summary', mode="w+")

        profile_summary.summarize_truncated_nfw_mass_profile(summary_file=summary_file, truncated_nfw=MockTruncatedNFW(),
                                                             critical_surface_mass_density_arcsec=1.0,
                                                             cosmic_average_mass_density_arcsec=1.0)

        summary_file.close()

        summary_file_read = open(test_summary_dir + 'model.summary', mode="r")
        summary_text = summary_file_read.readlines()

        assert summary_text == ['Rho at scale radius = 100.00\n',
                                'Delta concentration = 200.00\n',
                                'Concentration = 300.00\n',
                                'Radius at 200x cosmic average density = 400.00"\n',
                                'Mass at 200x cosmic average density = 500.00 solMass\n',
                                'Mass at truncation radius = 600.00 solMass\n']

        summary_file.close()

        os.remove(path=test_summary_dir + 'model.summary')

    def test__summarize_truncated_nfw_challenge_mass_profile(self):

        summary_file = open(file=test_summary_dir + 'model.summary', mode="w+")

        profile_summary.summarize_truncated_nfw_challenge_mass_profile(summary_file=summary_file,
                                                                       truncated_nfw_challenge=MockTruncatedNFW())

        summary_file.close()

        summary_file_read = open(test_summary_dir + 'model.summary', mode="r")
        summary_text = summary_file_read.readlines()

        assert summary_text == ['Rho at scale radius = 100.00\n',
                                'Delta concentration = 200.00\n',
                                'Concentration = 300.00\n',
                                'Radius at 200x cosmic average density = 400.00"\n',
                                'Mass at 200x cosmic average density = 500.00 solMass\n',
                                'Mass at truncation radius = 600.00 solMass\n']

        summary_file.close()

        os.remove(path=test_summary_dir + 'model.summary')